from .collect import format_styles, from_components, from_loaded
from .themes import Theme, get_default_theme, set_default_theme
from .types import CSSProperties, GlobalStyles

__all__ = (
    "CSSProperties",
    "GlobalStyles",
    "Theme",
    "format_styles",
    "from_components",
    "from_loaded",
    "get_default_theme",
    "set_default_theme",
)



from typing import (
    Annotated,
    Any,
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
)

_T = TypeVar("_T", covariant=True)


def get_element_generic_args(cls_or_obj: Any) -> tuple[type, ...] | None:
    """Get the generic arguments of the element class.

    Args:
        cls_or_obj (Any): The element to get the generic arguments of.

    Returns:
        dict[str, Any] | None: The generic arguments or :obj:`None`.
    """
    from ludic.base import BaseElement

    if hints := getattr(cls_or_obj, "__annotations__", None):
        children = hints.get("children")
        attrs = hints.get("attrs")
        if attrs and children:
            return (children, attrs)

    for base in getattr(cls_or_obj, "__orig_bases__", []):
        if issubclass(get_origin(base), BaseElement):
            return get_args(base)
    return None


def get_element_attrs_annotations(
    cls_or_obj: Any, include_extras: bool = False
) -> dict[str, Any]:
    """Get the annotations of the element.

    Args:
        cls_or_obj (type[Any]): The element to get the annotations of.
        include_extras (bool): Whether to include extra annotation info.

    Returns:
        dict[str, Any]: The attributes' annotations of the element.
    """
    if (args := get_element_generic_args(cls_or_obj)) is not None:
        return get_type_hints(args[-1], include_extras=include_extras)
    return {}


def get_annotations_metadata_of_type(
    annotations: dict[str, Any],
    expected_type: type[_T],
    default: _T | None = None,
) -> dict[str, _T]:
    """Get the metadata of the annotations with the given type.

    Args:
        annotations (dict[str, Any]): The annotations.
        expected_type (Any): The expected type.
        default (Any, optional): The default type.

    Returns:
        dict[str, Any]: The metadata.
    """
    result: dict[str, _T] = {}
    for name, annotation in annotations.items():
        if get_origin(annotation) is not Annotated:
            continue
        for metadata in annotation.__metadata__:
            if isinstance(metadata, expected_type):
                result[name] = metadata
                break
        else:
            if default is not None:
                result[name] = default
    return result



from dataclasses import field

from ludic.styles.themes import Colors, Theme
from ludic.styles.types import Color, GlobalStyles


class FooTheme(Theme):
    name: str = "foo"

    colors: Colors = field(
        default_factory=lambda: Colors(
            primary=Color("#0d6efd"),
            secondary=Color("#6c757d"),
            success=Color("#198754"),
            info=Color("#0dcaf0"),
            warning=Color("#ffc107"),
            danger=Color("#dc3545"),
            light=Color("#414549"),
            dark=Color("#f8f8f8"),
            white=Color("#000"),
            black=Color("#fff"),
        )
    )

    @property
    def styles(self) -> GlobalStyles:
        return {}


class BarTheme(Theme):
    name: str = "bar"

    @property
    def styles(self) -> GlobalStyles:
        return {}



from collections.abc import Iterable
from typing import override

from ludic.attrs import GlobalAttrs
from ludic.html import dd, dl, dt
from ludic.types import Component, PrimitiveChildren

from .utils import attr_to_camel


class Key(Component[PrimitiveChildren, GlobalAttrs]):
    """Simple component rendering as the HTML ``dt`` element."""

    @override
    def render(self) -> dt:
        return dt(*self.children, **self.attrs)


class Value(Component[PrimitiveChildren, GlobalAttrs]):
    """Simple component rendering as the HTML ``dd`` element."""

    @override
    def render(self) -> dd:
        return dd(*self.children, **self.attrs)


class PairsAttrs(GlobalAttrs, total=False):
    """Attributes of the component ``Pairs``."""

    items: Iterable[tuple[str, PrimitiveChildren]]


class Pairs(Component[Key | Value, PairsAttrs]):
    """Simple component rendering as the HTML ``dl`` element.

    Example usage:

        Pairs(
            Key("Name"),
            Value("John"),
            Key("Age"),
            Value(42),
        )

    The components accepts the ``items`` attribute, which allow the following usage:

        Pairs(
            items=[("name", "John"), ("age", 42)],
        )

    Or alternatively:

        Pairs(
            items={"name": "John", "age": 42}.items(),
        )
    """

    @override
    def render(self) -> dl:
        from_items: list[Key | Value] = []

        for key, value in self.attrs.get("items", ()):
            from_items.append(Key(attr_to_camel(key)))
            from_items.append(Value(value))

        return dl(*from_items, *self.children, **self.attrs_for(dl))



from starlette.exceptions import HTTPException


class ClientError(HTTPException):
    def __init__(
        self,
        detail: str = "Client Error",
        headers: dict[str, str] | None = None,
        status_code: int = 400,
    ) -> None:
        super().__init__(status_code=status_code, detail=detail, headers=headers)


class BadRequestError(ClientError):
    def __init__(
        self,
        detail: str = "Bad Request",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=400, detail=detail, headers=headers)


class UnauthorizedError(ClientError):
    def __init__(
        self,
        detail: str = "Unauthorized",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=401, detail=detail, headers=headers)


class PaymentRequiredError(ClientError):
    def __init__(
        self,
        detail: str = "Payment Required",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=402, detail=detail, headers=headers)


class ForbiddenError(ClientError):
    def __init__(
        self,
        detail: str = "Forbidden",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=403, detail=detail, headers=headers)


class NotFoundError(ClientError):
    def __init__(
        self,
        detail: str = "Not Found",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=404, detail=detail, headers=headers)


class MethodNotAllowedError(ClientError):
    def __init__(
        self,
        detail: str = "Method Not Allowed",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=405, detail=detail, headers=headers)


class TooManyRequestsError(ClientError):
    def __init__(
        self,
        detail: str = "Too Many Requests",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=429, detail=detail, headers=headers)


class ServerError(HTTPException):
    def __init__(
        self,
        detail: str = "Server Error",
        headers: dict[str, str] | None = None,
        status_code: int = 500,
    ) -> None:
        super().__init__(status_code=status_code, detail=detail, headers=headers)


class InternalServerError(ServerError):
    def __init__(
        self,
        detail: str = "Internal Server Error",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=500, detail=detail, headers=headers)


class NotImplementedError(ServerError):
    def __init__(
        self,
        detail: str = "Not Implemented",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=501, detail=detail, headers=headers)


class BadGatewayError(ServerError):
    def __init__(
        self,
        detail: str = "Bad Gateway",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=502, detail=detail, headers=headers)


class ServiceUnavailableError(ServerError):
    def __init__(
        self,
        detail: str = "Service Unavailable",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=503, detail=detail, headers=headers)


class GatewayTimeoutError(ServerError):
    def __init__(
        self,
        detail: str = "Gateway Timeout",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code=504, detail=detail, headers=headers)



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypeVar

from .types import Color, ColorRange, Size

if TYPE_CHECKING:
    from ludic.types import BaseElement

_T = TypeVar("_T", bound="BaseElement")


@dataclass
class Colors:
    """Colors for a theme."""

    primary: Color = ColorRange(["#276662", "#4ecdc4", "#dbf5f3"])
    secondary: Color = ColorRange(["#212529", "#414549", "#616569"])
    success: Color = ColorRange(["#637a32", "#c7f464", "#eefbd0"])
    info: Color = ColorRange(["#978801", "#fce303", "#fef9cc"])
    warning: Color = ColorRange(["#7e4801", "#fc9003", "#fee8cc"])
    danger: Color = ColorRange(["#711414", "#e32929", "#f9d4d4"])

    light: Color = Color("#f8f8f8")
    dark: Color = Color("#414549")

    white: Color = Color("#fff")
    black: Color = Color("#222")


@dataclass
class Header:
    """Header for a theme."""

    size: Size = Size(1.5, "em")
    anchor: bool = True


@dataclass
class Headers:
    """Headers for a theme."""

    h1: Header = field(default_factory=lambda: Header(size=Size(3, "em"), anchor=False))
    h2: Header = field(
        default_factory=lambda: Header(size=Size(2.5, "em"), anchor=False)
    )
    h3: Header = field(default_factory=lambda: Header(size=Size(2, "em"), anchor=False))
    h4: Header = field(
        default_factory=lambda: Header(size=Size(1.5, "em"), anchor=False)
    )
    h5: Header = field(
        default_factory=lambda: Header(size=Size(1.25, "em"), anchor=False)
    )
    h6: Header = field(default_factory=lambda: Header(size=Size(1, "em"), anchor=False))


@dataclass
class Fonts:
    """Font sizes for a theme."""

    plain: str = "Helvetica Neue, Helvetica, Arial, sans-serif"
    serif: str = "Georgia, serif"
    mono: str = "Space Mono, Roboto Mono, monospace"

    size: Size = Size(1, "em")


@dataclass
class Sizes:
    """Size for a theme."""

    xxxxs: Size = Size(0.39)
    xxxs: Size = Size(0.47)
    xxs: Size = Size(0.57)
    xs: Size = Size(0.69)
    s: Size = Size(0.84)
    m: Size = Size(1)
    l: Size = Size(1.16)  # noqa
    xl: Size = Size(1.4)
    xxl: Size = Size(1.68)
    xxxl: Size = Size(2.01)
    xxxxl: Size = Size(2.41)


@dataclass
class Borders:
    """Border sizes for a theme."""

    thin: Size = Size(0.1)
    normal: Size = Size(0.23)
    thick: Size = Size(0.42)


@dataclass
class Rounding:
    """Border rounding for a theme."""

    less: Size = Size(0.20)
    normal: Size = Size(0.25)
    more: Size = Size(0.35)


@dataclass
class Sidebar:
    """Sidebar layout config for a theme."""

    # The width of the sidebar (empty means not set; defaults to the content width)
    side_width: Size | None = None

    # The narrowest the content element can be before wrapping. Should be a percentage.
    content_min_width: str = "60%"


@dataclass
class Switcher:
    """Switcher layout config for a theme."""

    # The container width at which the component switches between a horizontal and
    # vertical layout
    threshold: Size = Size(40, "rem")

    # The maximum number of elements allowed to appear in the horizontal configuration
    limit: int = 4


@dataclass
class Cover:
    """Cover layout config for a theme."""

    # The minimum height of the cover
    min_height: Size = Size(100, "vh")

    # The minimum space between and around the child elements
    element: str = "h1"


@dataclass
class Grid:
    """Grid layout config for a theme."""

    # The size of the grid cells
    cell_size: Size = Size(250, "px")


@dataclass
class Frame:
    """Frame layout config for a theme."""

    # The width of the frame
    numerator: int = 16

    # The height of the frame
    denominator: int = 9


@dataclass
class Layouts:
    """Layout configuration for a theme."""

    sidebar: Sidebar = field(default_factory=Sidebar)
    switcher: Switcher = field(default_factory=Switcher)
    cover: Cover = field(default_factory=Cover)
    grid: Grid = field(default_factory=Grid)
    frame: Frame = field(default_factory=Frame)


@dataclass
class CodeBlock:
    """Code block config for a theme."""

    background_color: Color = Color("#f8f8f8")
    style: str | type = "default"


@dataclass
class Theme(metaclass=ABCMeta):
    """An abstract base class for theme configuration."""

    measure: Size = Size(70, "ch")
    line_height: float = 1.5

    fonts: Fonts = field(default_factory=Fonts)
    colors: Colors = field(default_factory=Colors)
    sizes: Sizes = field(default_factory=Sizes)

    borders: Borders = field(default_factory=Borders)
    rounding: Rounding = field(default_factory=Rounding)
    headers: Headers = field(default_factory=Headers)

    layouts: Layouts = field(default_factory=Layouts)
    code: CodeBlock = field(default_factory=CodeBlock)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Theme) and self.name == other.name

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the theme."""
        raise NotImplementedError

    def use(self, element: _T) -> _T:
        """Apply the theme to an element.

        Args:
            element (_T): Element to apply the theme to.

        Returns:
            _T: The element with the theme applied.
        """
        element.context["theme"] = self
        return element


@dataclass
class DarkTheme(Theme):
    """The dark theme."""

    name: str = "dark"

    colors: Colors = field(
        default_factory=lambda: Colors(
            primary=Color("#0d6efd"),
            secondary=Color("#6c757d"),
            success=Color("#198754"),
            info=Color("#0dcaf0"),
            warning=Color("#ffc107"),
            danger=Color("#dc3545"),
            light=Color("#414549"),
            dark=Color("#f8f8f8"),
            white=Color("#000"),
            black=Color("#fff"),
        )
    )


@dataclass
class LightTheme(Theme):
    """Light theme."""

    name: str = "light"


_DEFAULT_THEME: Theme = LightTheme()


def get_default_theme() -> Theme:
    """Get the default theme."""
    return _DEFAULT_THEME


def set_default_theme(theme: Theme) -> None:
    """Set the default theme."""
    global _DEFAULT_THEME
    _DEFAULT_THEME = theme



from typing import NotRequired, override

from ludic.attrs import GlobalAttrs
from ludic.html import div, style
from ludic.types import AnyChildren, Component


class Loading(Component[AnyChildren, GlobalAttrs]):
    classes = ["loader"]
    styles = style.use(
        lambda theme: {
            ".loader": {
                "text-align": "center",
            },
            ".lds-ellipsis": {
                "display": "inline-block",
                "position": "relative",
                "width": theme.sizes.m * 5,
                "height": theme.sizes.m,
                "margin": theme.sizes.xl,
            },
            ".lds-ellipsis div": {
                "position": "absolute",
                "width": theme.sizes.m,
                "height": theme.sizes.m,
                "border-radius": "50%",
                "background": theme.colors.dark,
                "animation-timing-function": "cubic-bezier(0, 1, 1, 0)",
            },
            ".lds-ellipsis div:nth-child(1)": {
                "left": "8px",
                "animation": "lds-ellipsis1 0.6s infinite",
            },
            ".lds-ellipsis div:nth-child(2)": {
                "left": "8px",
                "animation": "lds-ellipsis2 0.6s infinite",
            },
            ".lds-ellipsis div:nth-child(3)": {
                "left": "32px",
                "animation": "lds-ellipsis2 0.6s infinite",
            },
            ".lds-ellipsis div:nth-child(4)": {
                "left": "56px",
                "animation": "lds-ellipsis3 0.6s infinite",
            },
            "@keyframes lds-ellipsis1": {
                "0%": {
                    "transform": "scale(0)",
                },
                "100%": {
                    "transform": "scale(1)",
                },
            },
            "@keyframes lds-ellipsis3": {
                "0%": {
                    "transform": "scale(1)",
                },
                "100%": {
                    "transform": "scale(0)",
                },
            },
            "@keyframes lds-ellipsis2": {
                "0%": {
                    "transform": "translate(0, 0)",
                },
                "100%": {
                    "transform": "translate(24px, 0)",
                },
            },
        }
    )

    @override
    def render(self) -> div:
        return div(
            div(div(""), div(""), div(""), div(""), classes=["lds-ellipsis"]),
            **self.attrs_for(div),
        )


class LazyLoaderAttrs(GlobalAttrs):
    load_url: str
    placeholder: NotRequired[AnyChildren]


class LazyLoader(Component[AnyChildren, LazyLoaderAttrs]):
    """Lazy loader component using HTMX attributes.

    Usage:

        LazyLoader(
            load_url="https://example.com/svg-file"
        )
    """

    @override
    def render(self) -> div:
        self.attrs.setdefault("hx_trigger", "load")
        self.attrs.setdefault("hx_get", self.attrs["load_url"])
        self.attrs.setdefault("hx_swap", "outerHTML")
        return div(self.attrs.get("placeholder", Loading()), **self.attrs_for(div))



from ludic.html import (
    a,
    b,
    button,
    div,
    i,
    label,
    p,
    script,
    span,
    table,
    tbody,
    td,
    th,
    thead,
    tr,
)
from ludic.styles import CSSProperties


def test_empty_element() -> None:
    dom = div()
    assert dom.to_html() == "<div />"
    assert dom.to_string() == "<div />"

    dom2 = p("")
    assert dom2.to_html() == "<p></p>"
    assert dom2.to_string() == "<p></p>"

    dom3 = script()
    assert dom3.to_html() == "<script></script>"
    assert dom3.to_string() == "<script />"


def test_html_paragraph() -> None:
    paragraph = p(f"Hello, World! {b("Something bold")} and {i("Something italic")}")
    assert paragraph.to_html() == (
        "<p>Hello, World! <b>Something bold</b> and <i>Something italic</i></p>"
    )
    assert paragraph.to_string() == (
        "<p>\n"
        "  Hello, World! \n"
        "  <b>Something bold</b>\n"
        "   and \n"
        "  <i>Something italic</i>\n"
        "</p>"
    )  # fmt: skip


def test_html_link() -> None:
    link = a("A link!", href="https://example.com")
    assert link.to_html() == '<a href="https://example.com">A link!</a>'
    assert link.to_string() == (
        '<a href="https://example.com">\n'
        "  A link!\n"
        "</a>"
    )  # fmt: skip


def test_html_table() -> None:
    dom = table(
        thead(
            tr(
                th("Header 1"),
                th("Header 2"),
                th("Header 3"),
            )
        ),
        tbody(
            tr(
                td("Cell 1", style=CSSProperties(color="red", height="100px")),
                td("Cell 2"),
                td("Cell 3"),
            ),
        ),
    )

    assert dom.to_html() == (
        "<table>"
            "<thead>"
                "<tr>"
                    "<th>Header 1</th>"
                    "<th>Header 2</th>"
                    "<th>Header 3</th>"
                "</tr>"
            "</thead>"
            "<tbody>"
                "<tr>"
                    '<td style="color:red;height:100px">Cell 1</td>'
                    "<td>Cell 2</td>"
                    "<td>Cell 3</td>"
                "</tr>"
            "</tbody>"
        "</table>"
    )  # fmt: skip


def test_button_get() -> None:
    dom = div(
        div(label("First Name"), ": Joe"),
        div(label("Last Name"), ": Blow"),
        div(label("Email"), ": joe@blow.com"),
        button(
            "Click To Edit",
            classes=["btn", "btn-primary"],
            hx_get="/contact/1/edit",
        ),
        hx_target="this",
        hx_swap="outerHTML",
    )

    assert isinstance(dom.children[3], button)
    assert dom.children[3].attrs.get("hx_get") == "/contact/1/edit"
    assert dom.children[3].children[0] == "Click To Edit"
    assert dom.attrs.get("hx_target") == "this"
    assert dom.to_html() == (
        '<div hx-target="this" hx-swap="outerHTML">'
            "<div><label>First Name</label>: Joe</div>"
            "<div><label>Last Name</label>: Blow</div>"
            "<div><label>Email</label>: joe@blow.com</div>"
            '<button class="btn btn-primary" hx-get="/contact/1/edit">'
                "Click To Edit"
            "</button>"
        "</div>"
    )  # fmt: skip
    assert dom.to_string() == (
        '<div hx-target="this" hx-swap="outerHTML">\n'
        "  <div>\n"
        "    <label>First Name</label>\n"
        "    : Joe\n"
        "  </div>\n"
        "  <div>\n"
        "    <label>Last Name</label>\n"
        "    : Blow\n"
        "  </div>\n"
        "  <div>\n"
        "    <label>Email</label>\n"
        "    : joe@blow.com\n"
        "  </div>\n"
        '  <button class="btn btn-primary" hx-get="/contact/1/edit">\n'
        "    Click To Edit\n"
        "  </button>\n"
        "</div>"
    )


def test_expand_nested() -> None:
    div_dom = div(p(1), p(2), p(3))
    span_dom = span(*div_dom)

    assert div_dom.text == span_dom.text == "123"
    assert span_dom.to_html() == "<span><p>1</p><p>2</p><p>3</p></span>"


def test_repr_and_str() -> None:
    dom = div(p(1), p(2), p(3), id="test")
    assert repr(dom) == '<div id="test"><p>1</p><p>2</p><p>3</p></div>'
    assert str(dom) == (
        '<div id="test">\n'
        "  <p>1</p>\n"
        "  <p>2</p>\n"
        "  <p>3</p>\n"
        "</div>"
    )  # fmt: skip



from typing import override

from ludic.attrs import GlobalAttrs
from ludic.html import div, style
from ludic.types import AnyChildren, ComponentStrict


class Stack(div):
    """Stack is a block component which renders its children with a space between them.

    Example usage:

        Stack(
            h2("Header 2"),
            p("Hello, World!"),
            p("Goodbye, World!"),
        )

    All children components will have a space (margin) separating them.
    """

    classes = ["stack"]
    styles = style.use(
        lambda theme: {
            ".stack": {
                "display": "flex",
                "flex-direction": "column",
                "justify-content": "flex-start",
            },
            ".stack > *": {
                "margin-block": "0",
                "inline-size": "100%",
            },
            ".stack > * + *": {
                "margin-block-start": theme.sizes.xl,
            },
            ".stack.recursive *": {
                "margin-block": "0",
            },
            ".stack.recursive * + *": {
                "margin-block-start": theme.sizes.xl,
            },
            ".stack.tiny > * + *": {
                "margin-block-start": theme.sizes.xxxxs,
            },
            ".stack.small > * + *": {
                "margin-block-start": theme.sizes.s,
            },
            ".stack.large > * + *": {
                "margin-block-start": theme.sizes.xxxxl,
            },
        }
    )


class Box(div):
    """A component which applies padding from all sides.

    Thus the inner content of the `Box` is spaced in all directions equally.

    Example usage:

        Box(
            h2("List of Things"),
            ul(
                li("Thing 1"),
                li("Thing 2"),
            )
        )
    """

    classes = ["box"]
    styles = style.use(
        lambda theme: {
            ".box": {
                "padding": theme.sizes.l,
                "color": theme.colors.dark,
            },
            ".box.small": {
                "padding": theme.sizes.s,
            },
            ".box.large": {
                "padding": theme.sizes.xxl,
            },
            ".box:not(.transparent)": {
                "border": (
                    f"{theme.borders.thin} solid {theme.colors.light.darken(1)}"
                ),
                "border-radius": theme.rounding.more,
                "background-color": theme.colors.light,
            },
            ".box:not(.transparent) *": {
                "color": "inherit",
            },
            ".box.invert": {
                "color": theme.colors.light,
                "background-color": theme.colors.dark,
                "border": f"{theme.borders.thin} solid {theme.colors.dark}",
                "border-radius": theme.rounding.more,
            },
        }
    )


class Center(div):
    """A component which horizontally centers its children.

    Example usage:

        Center(
            h2("Lorem Ipsum"),
            p("Dolor sit amet..."),
        )
    """

    classes = ["center"]
    styles = style.use(
        lambda theme: {
            ".center": {
                "box-sizing": "content-box",
                "margin-inline": "auto",
                "max-inline-size": theme.measure,
                "padding-inline-start": theme.sizes.xxxl,
                "padding-inline-end": theme.sizes.xxxl,
            }
        }
    )


class Cluster(div):
    """A component for inline children to be rendered in a row.

    All contained children have a space (margin) between them.

    Example usage:

        Cluster(
            button("button 1"),
            button("button 2"),
        )
    """

    classes = ["cluster"]
    styles = style.use(
        lambda theme: {
            ".cluster": {
                "display": "flex",
                "flex-wrap": "wrap",
                "gap": theme.sizes.l,
                "justify-content": "flex-start",
                "align-items": "center",
            },
            ".cluster.centered": {
                "justify-content": "center",
            },
            ".cluster.flex-end": {
                "justify-content": "flex-end",
            },
            ".cluster.small": {
                "gap": theme.sizes.xxs,
            },
            ".cluster.large": {
                "gap": theme.sizes.xxl,
            },
        }
    )


class Sidebar(div):
    """The sidebar part of a WithSidebar component."""

    classes = ["sidebar"]


class WithSidebar(ComponentStrict[AnyChildren, AnyChildren, GlobalAttrs]):
    """A component with a content and a sidebar.

    Example usage:

        WithSidebar(
            Sidebar(...),
            div(...),
        )

    Or you can put the sidebar on the right side:

        WithSidebar(
            div(...),
            Sidebar(...),
        )
    """

    classes = ["with-sidebar"]
    styles = style.use(
        lambda theme: {
            ".with-sidebar": {
                "display": "flex",
                "flex-wrap": "wrap",
                "gap": theme.sizes.xxxxl,
            },
            ".with-sidebar > .sidebar": (
                {
                    "flex-basis": theme.layouts.sidebar.side_width,
                    "flex-grow": 1,
                }
                if theme.layouts.sidebar.side_width
                else {
                    "flex-grow": 1,
                }
            ),
            ".with-sidebar > :not(.sidebar)": {
                "flex-basis": 0,
                "flex-grow": 999,
                "min-inline-size": theme.layouts.sidebar.content_min_width,
            },
        }
    )

    @override
    def render(self) -> div:
        return div(self.children[0], self.children[1])


class Switcher(div):
    """A component switching between horizontal and vertical layouts.

    All the children are either composed horizontally or vertically
    depending on the width of the viewport.

    Example usage:

        Switcher(
            div(...),
            div(...),
            div(...),
        )
    """

    classes = ["switcher"]
    styles = style.use(
        lambda theme: {
            ".switcher": {
                "display": "flex",
                "flex-wrap": "wrap",
                "gap": theme.sizes.xxl,
            },
            ".switcher > *": {
                "flex-grow": 1,
                "flex-basis": (
                    f"calc(({theme.layouts.switcher.threshold} - 100%) * 999)"
                ),
            },
            (
                f".switcher > :nth-last-child(n+{theme.layouts.switcher.limit+1})",
                f".switcher > :nth-last-child(n+{theme.layouts.switcher.limit+1}) ~ *",
            ): {
                "flex-basis": "100%",
            },
        }
    )


class Cover(div):
    """A component which covers the whole viewport.

    Example usage:

        Cover(
            div(...),
        )
    """

    classes = ["cover"]
    styles = style.use(
        lambda theme: {
            ".cover": {
                "display": "flex",
                "flex-direction": "column",
                "min-block-size": theme.layouts.cover.min_height,
                "padding": theme.sizes.xxl,
            },
            ".cover > *": {
                "margin-block": theme.sizes.xxl,
            },
            f".cover > :first-child:not({theme.layouts.cover.element})": {
                "margin-block-start": "0",
            },
            f".cover > :last-child:not({theme.layouts.cover.element})": {
                "margin-block-end": "0",
            },
            f".cover > {theme.layouts.cover.element}": {
                "margin-block": "auto",
            },
        }
    )


class Grid(div):
    """A component which creates a grid layout.

    Example usage:

        Grid(
            div(...),
            div(...),
        )
    """

    classes = ["grid"]
    styles = style.use(
        lambda theme: {
            ".grid": {
                "display": "grid",
                "grid-gap": theme.sizes.xxl,
            },
            f"@supports (width: min({theme.layouts.grid.cell_size}, 100%))": {
                ".grid": {
                    "grid-template-columns": (
                        "repeat("
                        "auto-fit,"
                        f"minmax(min({theme.layouts.grid.cell_size},"
                        "100%"
                        "), 1fr))"
                    ),
                },
            },
        }
    )


class Frame(div):
    """A component which creates a frame layout.

    Example usage:

        Frame(
            div(...),
        )
    """

    classes = ["frame"]
    styles = style.use(
        lambda theme: {
            ".frame": {
                "aspect-ratio": (
                    f"{theme.layouts.frame.numerator}/{theme.layouts.frame.denominator}"
                ),
                "overflow": "hidden",
                "display": "flex",
                "justify-content": "center",
                "align-items": "center",
            },
            (".frame > img", ".frame > video"): {
                "inline-size": "100%",
                "block-size": "100%",
                "object-fit": "cover",
            },
        }
    )



from collections.abc import Mapping
from typing import Literal, LiteralString, Self, SupportsIndex, TypedDict

from .utils import (
    darken_color,
    hex_to_rgb,
    lighten_color,
    pick_readable_color_for,
)

SizeUnit = Literal["px", "ex", "em", "ch", "rem", "vw", "vh", "vmin", "vmax", "%"]


class Color(str):
    """Color class."""

    @property
    def rgb(self) -> tuple[int, int, int]:
        """RGB color."""
        return hex_to_rgb(self)

    def darken(self, shift: int = 1) -> Self:
        """Pick darker color from the range by given shift.

        Args:
            shift (int, optional): Darkening shift. Defaults to 1.

        Returns:
            str: Darkened color.
        """
        return type(self)(darken_color(self, shift / 20))

    def lighten(self, shift: int = 1) -> Self:
        """Pick lighter color from the range by given shift.

        Args:
            shift (int, optional): Lightening shift. Defaults to 1.

        Returns:
            str: Lightened color.
        """
        return type(self)(lighten_color(self, shift / 20))

    def readable(self) -> Self:
        """Get lighter or darker variant of the given color depending on the luminance.

        Args:
            color (str): Color to find the readable opposite for.

        Returns:
            str: Readable opposite of the given color.
        """
        return type(self)(pick_readable_color_for(self))


class ColorRange(Color):
    """Color class."""

    variants: list[str]
    position: int

    def __new__(
        cls, variants: list[str] | str, position: int | None = None
    ) -> "ColorRange":
        if isinstance(variants, str):
            variants = [variants]

        new_position = position or (len(variants) // 2)

        self = super().__new__(cls, variants[new_position])
        self.variants = variants
        self.position = new_position
        return self

    def darken(self, shift: int = 1) -> Self:
        """Pick darker color from the range by given shift.

        Args:
            shift (int, optional): Darkening shift. Defaults to 1.

        Returns:
            str: Darkened color.
        """
        if 0 <= self.position - shift < len(self.variants):
            return type(self)(self.variants, self.position - shift)
        else:
            return type(self)(self.variants, 0)

    def lighten(self, shift: int = 1) -> Self:
        """Pick lighter color from the range by given shift.

        Args:
            shift (int, optional): Lightening shift. Defaults to 1.

        Returns:
            str: Lightened color.
        """
        if 0 <= self.position + shift < len(self.variants):
            return type(self)(self.variants, self.position + shift)
        else:
            return type(self)(self.variants, len(self.variants) - 1)


class Size(str):
    """Size class."""

    value: float
    unit: SizeUnit = "rem"

    def __new__(cls, value: float, unit: SizeUnit = "rem") -> "Size":
        match unit:
            case "px" | "ch":
                self = super().__new__(cls, f"{value:.0f}{unit}")
            case _:
                formatted = f"{value:.2f}".rstrip("0").rstrip(".")
                if not formatted or formatted == "0":
                    formatted = "0"
                else:
                    formatted = f"{formatted}{unit}"
                self = super().__new__(cls, formatted)

        self.value = value
        self.unit = unit
        return self

    def __mul__(self, factor: float | int | LiteralString | SupportsIndex) -> Self:
        """Scale size by a given factor.

        Args:
            factor (float): Scaling factor.

        Returns:
            str: Scaled size.
        """
        if isinstance(factor, float | int):
            return type(self)(self.value * factor, self.unit)
        else:
            return self

    def __add__(self, value: float | int | LiteralString | SupportsIndex) -> Self:
        """Increment size by a given factor.

        Args:
            value (float, optional): Increment value.

        Returns:
            str: Incremented size.
        """
        if isinstance(value, float | int):
            return type(self)(self.value + value, self.unit)
        else:
            return self

    def __sub__(self, value: float | int | LiteralString | SupportsIndex) -> Self:
        """Decrement size by a given factor.

        Args:
            value (float, optional): Decrement value.

        Returns:
            str: Decremented size.
        """
        if isinstance(value, float | int):
            return type(self)(self.value - value, self.unit)
        else:
            return self


CSSProperties = TypedDict(
    "CSSProperties",
    {
        # A
        "align-content": Literal[
            "center",
            "flex-start",
            "flex-end",
            "space-between",
            "space-around",
            "stretch",
        ],
        "align-items": Literal[
            "stretch", "center", "flex-start", "flex-end", "baseline"
        ],
        "align-self": Literal[
            "auto",
            "stretch",
            "center",
            "flex-start",
            "flex-end",
            "baseline",
            "initial",
            "inherit",
        ],
        "all": str,
        "animation": str,
        "animation-delay": str,
        "animation-direction": Literal[
            "normal", "reverse", "alternate", "alternate-reverse"
        ],
        "animation-duration": str,
        "animation-fill-mode": Literal["none", "forwards", "backwards", "both"],
        "animation-iteration-count": float,
        "animation-name": str,
        "animation-play-state": Literal["paused", "running"],
        "animation-timing-function": str,
        "aspect-ratio": str,
        # B
        "backdrop-filter": str,
        "backface-visibility": Literal["visible", "hidden"],
        "background": str,
        "background-attachment": Literal["scroll", "fixed", "local"],
        "background-blend-mode": str,
        "background-clip": Literal["border-box", "padding-box", "content-box", "text"],
        "background-color": str,
        "background-image": str,
        "background-origin": Literal["padding-box", "border-box", "content-box"],
        "background-position": str,
        "background-repeat": Literal[
            "repeat", "repeat-x", "repeat-y", "no-repeat", "space", "round"
        ],
        "background-size": Literal["auto", "cover", "contain"],
        "block-size": str,
        "border": str,
        "border-block": str,
        "border-block-color": str,
        "border-block-end": str,
        "border-block-end-color": str,
        "border-block-end-style": str,
        "border-block-end-width": str,
        "border-block-start": str,
        "border-block-start-color": str,
        "border-block-start-style": str,
        "border-block-start-width": str,
        "border-block-style": str,
        "border-block-width": str,
        "border-bottom": str,
        "border-bottom-color": str,
        "border-bottom-left-radius": str,
        "border-bottom-right-radius": str,
        "border-bottom-style": str,
        "border-bottom-width": str,
        "border-collapse": Literal["collapse", "separate"],
        "border-color": str,
        "border-image": str,
        "border-image-outset": str,
        "border-image-repeat": Literal["stretch", "repeat", "round"],
        "border-image-slice": str,
        "border-image-source": str,
        "border-image-width": str,
        "border-inline": str,
        "border-inline-color": str,
        "border-inline-end": str,
        "border-inline-end-color": str,
        "border-inline-end-style": str,
        "border-inline-end-width": str,
        "border-inline-start": str,
        "border-inline-start-color": str,
        "border-inline-start-style": str,
        "border-inline-start-width": str,
        "border-inline-style": str,
        "border-inline-width": str,
        "border-left": str,
        "border-left-color": str,
        "border-left-style": str,
        "border-left-width": str,
        "border-radius": str,
        "border-right": str,
        "border-right-color": str,
        "border-right-style": str,
        "border-right-width": str,
        "border-spacing": str,
        "border-style": str,
        "border-top": str,
        "border-top-color": str,
        "border-top-left-radius": str,
        "border-top-right-radius": str,
        "border-top-style": str,
        "border-top-width": str,
        "border-width": str,
        "bottom": str,
        "box-decoration-break": Literal["slice", "clone"],
        "box-shadow": str,
        "box-sizing": Literal["content-box", "border-box"],
        # C
        "caption-side": Literal["top", "bottom"],
        "caret-color": str,
        "clear": Literal["none", "left", "right", "both"],
        "clip": str,
        "clip-path": str,
        "color": str,
        "column-count": str,
        "column-fill": Literal["auto", "balance"],
        "column-gap": str,
        "column-rule": str,
        "column-rule-color": str,
        "column-rule-style": str,
        "column-rule-width": str,
        "column-span": Literal["none", "all"],
        "column-width": str,
        "columns": str,
        "content": str,
        "counter-increment": str,
        "counter-reset": str,
        "cursor": Literal[
            "alias",
            "all-scroll",
            "auto",
            "cell",
            "context-menu",
            "col-resize",
            "copy",
            "crosshair",
            "default",
            "e-resize",
            "ew-resize",
            "grab",
            "grabbing",
            "help",
            "move",
            "n-resize",
            "ne-resize",
            "nesw-resize",
            "ns-resize",
            "nw-resize",
            "nwse-resize",
            "no-drop",
            "none",
            "not-allowed",
            "pointer",
            "progress",
            "row-resize",
            "s-resize",
            "se-resize",
            "sw-resize",
            "text",
            "unset",
            "vertical-text",
            "w-resize",
            "wait",
            "zoom-in",
            "zoom-out",
        ],
        # D
        "direction": Literal["ltr", "rtl"],
        "display": Literal[
            "block",
            "contents",
            "flex",
            "grid",
            "inline",
            "inline-block",
            "inline-flex",
            "inline-grid",
            "inline-table",
            "list-item",
            "run-in",
            "table",
            "table-caption",
            "table-cell",
            "table-column",
            "table-column-group",
            "table-footer-group",
            "table-header-group",
            "table-row",
            "table-row-group",
            "none",
            "unset",
        ],
        "empty-cells": Literal["show", "hide"],
        # F
        "filter": str,
        "flex": str,
        "flex-basis": float | str,
        "flex-direction": Literal["row", "row-reverse", "column", "column-reverse"],
        "flex-flow": str,
        "flex-grow": float | str,
        "flex-shrink": float | str,
        "flex-wrap": Literal["nowrap", "wrap", "wrap-reverse"],
        "float": Literal["left", "right", "none"],
        "font": str,
        "font-family": str,
        "font-feature-settings": str,
        "font-kerning": Literal["auto", "normal", "none"],
        "font-language-override": str,
        "font-size": str,
        "font-size-adjust": str,
        "font-stretch": Literal[
            "normal",
            "ultra-condensed",
            "extra-condensed",
            "condensed",
            "semi-condensed",
            "semi-expanded",
            "expanded",
            "extra-expanded",
            "ultra-expanded",
        ],
        "font-style": Literal["normal", "italic", "oblique"],
        "font-synthesis": Literal["none", "weight", "style"],
        "font-variant": Literal["normal", "small-caps"],
        "font-variant-caps": Literal[
            "normal",
            "small-caps",
            "all-small-caps",
            "petite-caps",
            "all-petite-caps",
            "unicase",
            "titling-caps",
        ],
        "font-variant-east-asian": Literal[
            "normal",
            "ruby",
            "jis78",
            "jis83",
            "jis90",
            "jis04",
            "simplified",
            "traditional",
        ],
        "font-variant-ligatures": Literal[
            "normal",
            "none",
            "common-ligatures",
            "no-common-ligatures",
            "discretionary-ligatures",
            "no-discretionary-ligatures",
            "historical-ligatures",
            "no-historical-ligatures",
            "contextual",
            "no-contextual",
        ],
        "font-variant-numeric": Literal[
            "normal",
            "ordinal",
            "slashed-zero",
            "lining-nums",
            "oldstyle-nums",
            "proportional-nums",
            "tabular-nums",
            "diagonal-fractions",
            "stacked-fractions",
        ],
        "font-variant-position": Literal["normal", "sub", "super"],
        "font-weight": Literal[
            "normal",
            "bold",
            "bolder",
            "lighter",
            "100",
            "200",
            "300",
            "400",
            "500",
            "600",
            "700",
            "800",
            "900",
        ],
        # G
        "gap": str,
        "grid": str,
        "grid-area": str,
        "grid-auto-columns": str,
        "grid-auto-flow": Literal["row", "column", "dense"],
        "grid-auto-rows": str,
        "grid-column": str,
        "grid-column-end": str,
        "grid-column-gap": str,
        "grid-column-start": str,
        "grid-gap": str,
        "grid-row": str,
        "grid-row-end": str,
        "grid-row-gap": str,
        "grid-row-start": str,
        "grid-template": str,
        "grid-template-areas": str,
        "grid-template-columns": str,
        "grid-template-rows": str,
        # H
        "hanging-punctuation": Literal[
            "none", "first", "last", "allow-end", "force-end"
        ],
        "height": str,
        "hyphens": Literal["none", "manual", "auto"],
        # I
        "inline-size": str,
        # J
        "justify-content": Literal[
            "flex-start",
            "flex-end",
            "center",
            "space-between",
            "space-around",
            "space-evenly",
            "start",
            "end",
            "left",
            "right",
        ],
        # L
        "left": str,
        "letter-spacing": str,
        "line-break": Literal["auto", "loose", "normal", "strict"],
        "line-height": float,
        "list-style": str,
        "list-style-image": str,
        "list-style-position": Literal["inside", "outside"],
        "list-style-type": Literal[
            "disc",
            "circle",
            "square",
            "decimal",
            "decimal-leading-zero",
            "lower-roman",
            "upper-roman",
            "lower-greek",
            "lower-alpha",
            "lower-latin",
            "upper-alpha",
            "upper-latin",
            "hebrew",
            "armenian",
            "georgian",
            "cjk-ideographic",
            "hiragana",
            "katakana",
            "hiragana-iroha",
            "katakana-iroha",
            "none",
        ],
        # M
        "margin": str,
        "margin-block": str,
        "margin-block-end": str,
        "margin-block-start": str,
        "margin-bottom": str,
        "margin-inline": str,
        "margin-inline-end": str,
        "margin-inline-start": str,
        "margin-left": str,
        "margin-right": str,
        "margin-top": str,
        "mask": str,
        "mask-border": str,
        "mask-border-mode": Literal["alpha", "luminance"],
        "mask-border-outset": str,
        "mask-border-repeat": str,
        "mask-border-slice": str,
        "mask-border-source": str,
        "mask-border-width": str,
        "mask-clip": str,
        "mask-composite": str,
        "mask-image": str,
        "mask-mode": str,
        "mask-origin": str,
        "mask-position": str,
        "mask-repeat": str,
        "mask-size": str,
        "mask-type": Literal["alpha", "luminance"],
        "max-block-size": str,
        "max-height": str,
        "max-inline-size": str,
        "max-width": str,
        "min-block-size": str,
        "min-height": str,
        "min-inline-size": str,
        "min-width": str,
        # O
        "object-fit": Literal["fill", "contain", "cover", "none", "scale-down"],
        "object-position": str,
        "offset": str,
        "offset-anchor": str,
        "offset-distance": str,
        "offset-path": str,
        "offset-position": str,
        "offset-rotate": str,
        "opacity": str,
        "order": str,
        "outline": str,
        "outline-color": str,
        "outline-offset": str,
        "outline-style": str,
        "outline-width": str,
        "overflow": Literal["auto", "hidden", "scroll", "visible"],
        "overflow-anchor": Literal["auto", "none"],
        "overflow-block": Literal["auto", "scroll", "visible"],
        "overflow-inline": Literal["auto", "scroll", "visible"],
        "overflow-wrap": Literal["normal", "break-word", "anywhere"],
        "overflow-x": Literal["auto", "hidden", "scroll", "visible"],
        "overflow-y": Literal["auto", "hidden", "scroll", "visible"],
        # P
        "padding": str,
        "padding-block": str,
        "padding-block-end": str,
        "padding-block-start": str,
        "padding-bottom": str,
        "padding-inline": str,
        "padding-inline-end": str,
        "padding-inline-start": str,
        "padding-left": str,
        "padding-right": str,
        "padding-top": str,
        "page-break-after": Literal["auto", "always", "avoid", "left", "right"],
        "page-break-before": Literal["auto", "always", "avoid", "left", "right"],
        "page-break-inside": Literal["auto", "avoid"],
        "paint-order": Literal["normal", "fill", "stroke", "markers"],
        "perspective": str,
        "perspective-origin": str,
        "place-content": str,
        "place-items": str,
        "place-self": str,
        "pointer-events": Literal[
            "auto",
            "none",
            "visiblePainted",
            "visibleFill",
            "visibleStroke",
            "visible",
            "painted",
            "fill",
            "stroke",
            "all",
        ],
        "position": Literal["static", "relative", "absolute", "fixed", "sticky"],
        # Q
        "quotes": str,
        # R
        "resize": Literal["none", "both", "horizontal", "vertical"],
        "right": str,
        # S
        "scroll-behavior": Literal["auto", "smooth"],
        "scroll-margin": str,
        "scroll-margin-block": str,
        "scroll-margin-block-end": str,
        "scroll-margin-block-start": str,
        "scroll-margin-bottom": str,
        "scroll-margin-inline": str,
        "scroll-margin-inline-end": str,
        "scroll-margin-inline-start": str,
        "scroll-margin-left": str,
        "scroll-margin-right": str,
        "scroll-margin-top": str,
        "scroll-padding": str,
        "scroll-padding-block": str,
        "scroll-padding-block-end": str,
        "scroll-padding-block-start": str,
        "scroll-padding-bottom": str,
        "scroll-padding-inline": str,
        "scroll-padding-inline-end": str,
        "scroll-padding-inline-start": str,
        "scroll-padding-left": str,
        "scroll-padding-right": str,
        "scroll-padding-top": str,
        "scroll-snap-align": Literal[
            "none",
            "start",
            "end",
            "center",
            "start end",
            "start center",
            "end center",
            "inherit",
            "initial",
            "unset",
        ],
        "scroll-snap-stop": Literal["normal", "always"],
        "scroll-snap-type": Literal[
            "none",
            "x",
            "y",
            "block",
            "inline",
            "both",
            "start",
            "end",
            "inherit",
            "initial",
            "unset",
        ],
        "shape-image-threshold": str,
        "shape-margin": str,
        "shape-outside": str,
        "tab-size": str,
        "table-layout": Literal["auto", "fixed"],
        "text-align": Literal[
            "left",
            "right",
            "center",
            "justify",
            "start",
            "end",
            "match-parent",
            "justify-all",
        ],
        "text-align-last": Literal["auto", "left", "right", "center", "justify"],
        "text-combine-upright": Literal["none", "all", "digits"],
        "text-decoration": str,
        "text-decoration-color": str,
        "text-decoration-line": Literal[
            "none", "underline", "overline", "line-through"
        ],
        "text-decoration-style": Literal["solid", "double", "dotted", "dashed", "wavy"],
        "text-emphasis": str,
        "text-emphasis-color": str,
        "text-emphasis-position": str,
        "text-emphasis-style": str,
        "text-indent": str,
        "text-justify": Literal[
            "auto",
            "inter-word",
            "inter-character",
            "none",
            "distribute",
            "distribute-all-lines",
            "inter-word",
        ],
        "text-orientation": Literal[
            "mixed", "upright", "sideways", "sideways-right", "use-glyph-orientation"
        ],
        "text-overflow": Literal["clip", "ellipsis"],
        "text-rendering": Literal[
            "auto", "optimizeSpeed", "optimizeLegibility", "geometricPrecision"
        ],
        "text-shadow": str,
        "text-transform": Literal[
            "none", "capitalize", "uppercase", "lowercase", "full-width"
        ],
        "text-underline-offset": str,
        "text-underline-position": str,
        "top": str,
        "transform": str,
        "transform-box": Literal["border-box", "fill-box", "view-box"],
        "transform-origin": str,
        "transform-style": Literal["flat", "preserve-3d"],
        "transition": str,
        "transition-delay": str,
        "transition-duration": str,
        "transition-property": Literal["all", "none"],
        "transition-timing-function": str,
        "translate": str,
        # U
        "unicode-bidi": Literal[
            "normal",
            "embed",
            "bidi-override",
            "isolate",
            "isolate-override",
            "plaintext",
        ],
        "user-select": Literal["auto", "text", "none", "contain", "all"],
        # V
        "vertical-align": Literal[
            "baseline",
            "sub",
            "super",
            "top",
            "text-top",
            "middle",
            "bottom",
            "text-bottom",
            "unset",
        ],
        "visibility": Literal["visible", "hidden", "collapse"],
        # W
        "white-space": Literal["normal", "nowrap", "pre", "pre-line", "pre-wrap"],
        "widows": str,
        "width": str,
        "will-change": str,
        "word-break": Literal["normal", "break-all", "keep-all"],
        "word-spacing": str,
        "word-wrap": Literal["normal", "break-word"],
        "writing-mode": Literal["horizontal-tb", "vertical-rl", "vertical-lr"],
        # Z
        "z-index": str,
    },
    total=False,
)

# FIXME: Currently, it is impossible to properly type nested CSS properties
# defined similar as in SCSS, it will be possible when the following PEP is
# implemented and supported by type checkers: https://peps.python.org/pep-0728/
GlobalStyles = Mapping[str | tuple[str, ...], "CSSProperties | GlobalStyles"]
"""CSS styles for elements or components which are defined by setting the ``styles``
class property.

Example usage:

    class Page(Component[AnyChildren, NoAttrs]):
        styles = {
            "body": {
                "background-color": "red",
            },
        }
"""



import colorsys


def hex_to_rgb(hex_color: str) -> tuple[int, int, int]:
    """Convert hex color to RGB.

    Args:
        hex_color (str): Hex color to convert.

    Returns:
        tuple[int, int, int]: RGB color.
    """
    if hex_color.startswith("#"):
        hex_color = hex_color[1:]

    if len(hex_color) == 3:
        hex_color = (
            f"{hex_color[0]}{hex_color[0]}"
            f"{hex_color[1]}{hex_color[1]}"
            f"{hex_color[2]}{hex_color[2]}"
        )

    if len(hex_color) != 6:
        raise ValueError(f"Invalid hex color: {hex_color}")

    return int(hex_color[:2], 16), int(hex_color[2:4], 16), int(hex_color[4:6], 16)


def rgb_to_hex(rgb: tuple[float, float, float]) -> str:
    """Convert RGB color to hex.

    Args:
        rgb (tuple[float, float, float]): RGB color to convert.

    Returns:
        str: Hex color.
    """
    return f"#{int(rgb[0]):02x}{int(rgb[1]):02x}{int(rgb[2]):02x}"


def pick_readable_color_for(color: str) -> str:
    """Get lighter or darker color variant of the given one depending on the luminance.

    Args:
        color (str): Color to find the readable opposite for.

    Returns:
        str: Readable opposite of the given color.
    """
    rgb = color if isinstance(color, tuple) else hex_to_rgb(color)
    return "#000" if sum(rgb) > 400 else "#fff"


def scale_color(color: str, factor: float = 0.5) -> str:
    """Scale a color by a given factor.

    Args:
        color (str): Color to scale.
        factor (float, optional): Scaling factor. Defaults to 0.5.

    Returns:
        str: Scaled color.
    """
    rgb = color if isinstance(color, tuple) else hex_to_rgb(color)
    hue, luminance, saturation = colorsys.rgb_to_hls(*rgb)

    if factor < 1:
        new_luminance = luminance * factor
    else:
        new_luminance = luminance + (255 - luminance) * (factor - 1)

    result = colorsys.hls_to_rgb(hue, min(255, max(1, new_luminance)), saturation)
    return rgb_to_hex(result)


def darken_color(color: str, factor: float = 0.5) -> str:
    """Darken a color by a given factor.

    Args:
        color (str): Color to darken.
        factor (float, optional): Darkening factor. Defaults to 0.5.

    Returns:
        : Darkened color.
    """
    return scale_color(color, min(1, max(0, 1 - factor)))


def lighten_color(color: str, factor: float = 0.5) -> str:
    """Lighten a color by a given factor.

    Args:
        color (str): Color to lighten.
        factor (float, optional): Lightening factor. Defaults to 0.5.

    Returns:
        : Lightened color.
    """
    return scale_color(color, max(1, min(2, 1 + factor)))


def first_not_none(*values: float | int | None) -> float | int:
    return next((value for value in values if value is not None), 0)


def clamp(min: str, val: str, max: str) -> str:
    return f"clamp({min}, {val}, {max})"



from typing import Self, override

from examples import Page

from ludic.catalog.headers import H1, H2
from ludic.catalog.quotes import Quote
from ludic.catalog.tables import Table, TableHead, TableRow
from ludic.types import Attrs, Blank, Component
from ludic.web import Endpoint, LudicApp
from ludic.web.datastructures import QueryParams

app = LudicApp(debug=True)


class ContactAttrs(Attrs):
    id: str
    name: str
    email: str


class ContactsSliceAttrs(Attrs):
    page: int
    contacts: list[ContactAttrs]


def load_contacts(page: int) -> list[ContactAttrs]:
    return [
        ContactAttrs(
            id=str(page * 10 + idx),
            email=f"void{page * 10 + idx}@null.org",
            name="Agent Smith",
        )
        for idx in range(10)
    ]


@app.get("/")
async def index() -> Page:
    return Page(
        H1("Infinite Scroll"),
        Quote(
            "The infinite scroll pattern provides a way to load content dynamically"
            "on user scrolling action.",
            source_url="https://htmx.org/examples/infinite-scroll/",
        ),
        H2("Demo"),
        ContactsTable(await ContactsSlice.get(QueryParams(page=1))),
    )


@app.endpoint("/contacts/")
class ContactsSlice(Endpoint[ContactsSliceAttrs]):
    @classmethod
    async def get(cls, params: QueryParams) -> Self:
        page = int(params.get("page", 1))
        return cls(page=page, contacts=load_contacts(page))

    @override
    def render(self) -> Blank[TableRow]:
        *init, last = (
            (contact["id"], contact["name"], contact["email"])
            for contact in self.attrs["contacts"]
        )
        return Blank(
            *(TableRow(*rows) for rows in init),
            TableRow(
                *last,
                hx_get=self.url_for(ContactsSlice).include_query_params(
                    page=self.attrs["page"] + 1
                ),
                hx_trigger="revealed",
                hx_swap="afterend",
            ),
        )


class ContactsTable(Component[ContactsSlice, Attrs]):
    @override
    def render(self) -> Table[TableHead, ContactsSlice]:
        return Table(
            TableHead("ID", "Name", "Email"),
            *self.children,
            classes=["text-align-center"],
        )



from ludic.web.requests import join_mounts


def test_join_mounts() -> None:
    assert join_mounts("", "") == ""
    assert join_mounts("a", "b") == "a:b"
    assert join_mounts("a:b", "c") == "a:b:c"
    assert join_mounts("foo:bar", "bar:Baz") == "foo:bar:Baz"
    assert join_mounts("foo:bar:Baz", "bar:Baz:foo") == "foo:bar:Baz:foo"



import pytest
from starlette.testclient import TestClient

from ludic.html import p
from ludic.web import LudicApp, Request
from ludic.web.exceptions import (
    BadGatewayError,
    BadRequestError,
    ClientError,
    ForbiddenError,
    GatewayTimeoutError,
    InternalServerError,
    MethodNotAllowedError,
    NotFoundError,
    NotImplementedError,
    PaymentRequiredError,
    ServerError,
    ServiceUnavailableError,
    TooManyRequestsError,
    UnauthorizedError,
)

app = LudicApp()


@app.get("/")
async def index() -> p:
    return p("hello world")


@app.get("/error")
async def error() -> p:
    raise RuntimeError("error happened")


@app.post("/only-post")
async def only_post() -> p:
    return p("only post")


@app.get("/not-found")
async def page_not_found() -> p:
    raise NotFoundError("This page does not exist.")


@app.exception_handler(404)
async def not_found(request: Request) -> p:
    return p(f"page {request.url.path} does not exist")


@app.exception_handler(405)
async def method_not_allowed() -> p:
    return p("method not allowed")


@app.exception_handler(500)
async def server_error(exception: Exception) -> p:
    return p(f"server error: {exception}")


def test_exception_handling() -> None:
    test_data = [
        ("/", 200, "hello world"),
        ("/not-found", 404, "page /not-found does not exist"),
        ("/only-post", 405, "method not allowed"),
        ("/error", 500, "server error: error happened"),
    ]

    with TestClient(app, raise_server_exceptions=False) as client:
        for path, status, content in test_data:
            response = client.get(path)
            assert response.status_code == status
            assert response.text == p(content).to_html()


def test_exceptions() -> None:
    test_data = [
        (ClientError, 400),
        (BadRequestError, 400),
        (UnauthorizedError, 401),
        (PaymentRequiredError, 402),
        (ForbiddenError, 403),
        (NotFoundError, 404),
        (MethodNotAllowedError, 405),
        (TooManyRequestsError, 429),
        (ServerError, 500),
        (InternalServerError, 500),
        (NotImplementedError, 501),
        (BadGatewayError, 502),
        (ServiceUnavailableError, 503),
        (GatewayTimeoutError, 504),
    ]

    for Error, status_code in test_data:
        assert issubclass(Error, Exception)
        with pytest.raises(Error) as error:
            raise Error("test message")

        assert (
            error.exconly()
            == f"ludic.web.exceptions.{Error.__name__}: {status_code}: test message"
        )



from typing import Annotated, Self, override

from examples import Page, init_db

from ludic.catalog.buttons import ButtonPrimary
from ludic.catalog.forms import FieldMeta, Form
from ludic.catalog.headers import H1, H2
from ludic.catalog.layouts import Cluster
from ludic.catalog.quotes import Quote
from ludic.catalog.tables import ColumnMeta, Table, create_rows
from ludic.html import span, style
from ludic.types import Attrs
from ludic.web import Endpoint, LudicApp
from ludic.web.parsers import ListParser

db = init_db()
app = LudicApp(debug=True)


class PersonAttrs(Attrs, total=False):
    id: Annotated[str, ColumnMeta(identifier=True)]
    name: Annotated[str, ColumnMeta()]
    email: Annotated[str, ColumnMeta()]
    active: Annotated[
        bool,
        ColumnMeta(kind=FieldMeta(kind="checkbox", label=None)),
    ]


class PeopleAttrs(Attrs):
    people: list[PersonAttrs]


class Toast(span):
    id: str = "toast"
    target: str = f"#{id}"
    styles = style.use(
        lambda theme: {
            Toast.target: {
                "background": theme.colors.success,
                "padding": f"{theme.sizes.xxxxs} {theme.sizes.xxxs}",
                "font-size": theme.fonts.size * 0.9,
                "border-radius": "3px",
                "opacity": "0",
                "transition": "opacity 3s ease-out",
            },
            f"{Toast.target}.htmx-settling": {
                "opacity": "100",
            },
        }
    )

    @override
    def render(self) -> span:
        return span(*self.children, id=self.id)


@app.get("/")
async def index() -> Page:
    return Page(
        H1("Bulk Update"),
        Quote(
            "This demo shows how to implement a common pattern where rows are "
            "selected and then bulk updated.",
            source_url="https://htmx.org/examples/bulk-update/",
        ),
        H2("Demo"),
        await PeopleTable.get(),
    )


@app.endpoint("/people/")
class PeopleTable(Endpoint[PeopleAttrs]):
    @classmethod
    async def post(cls, data: ListParser[PersonAttrs]) -> Toast:
        items = {row["id"]: row for row in data.validate()}
        activations = {True: 0, False: 0}

        for person in db.people.values():
            active = items.get(person.id, {}).get("active", False)
            if person.active != active:
                person.active = active
                activations[active] += 1

        return Toast(f"Activated {activations[True]}, deactivated {activations[False]}")

    @classmethod
    async def get(cls) -> Self:
        return cls(people=[person.dict() for person in db.people.values()])

    @override
    def render(self) -> Form:
        return Form(
            Table(*create_rows(self.attrs["people"], spec=PersonAttrs)),
            Cluster(
                ButtonPrimary("Bulk Update", type="submit"),
                Toast(),
            ),
            hx_post=self.url_for(PeopleTable),
            hx_target=Toast.target,
            hx_swap="outerHTML settle:3s",
        )



import inspect
from collections.abc import Callable, Mapping, Sequence
from functools import wraps
from typing import Any, Literal, TypeVar, cast

from starlette._utils import is_async_callable
from starlette.applications import AppType, Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import BaseRoute
from starlette.types import Lifespan
from starlette.websockets import WebSocket

from ludic.attrs import Attrs
from ludic.types import BaseElement

from .datastructures import URLPath
from .endpoints import Endpoint
from .responses import LudicResponse, run_in_threadpool_safe
from .routing import Router

TCallable = TypeVar("TCallable", bound=Callable[..., Any])
TEndpoint = TypeVar("TEndpoint", bound=Endpoint[Attrs])


class LudicApp(Starlette):
    """Starlette application with Ludic adoption.

    Example:

        async def homepage(request: Request) -> html:
            return html(...)

        app = LudicApp(debug=True)
        app.add_route("/", homepage)

    You can also use a method decorator to register a Ludic endpoint:

        @app.get("/")
        async def homepage(request: Request) -> button:
            return button(...)
    """

    router: Router

    def __init__(
        self,
        debug: bool = False,
        routes: Sequence[BaseRoute] | None = None,
        middleware: Sequence[Middleware] | None = None,
        exception_handlers: Mapping[Any, TCallable] | None = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: Lifespan[AppType] | None = None,
    ) -> None:
        super().__init__(debug, middleware=middleware)

        for key, value in (exception_handlers or {}).items():
            self.add_exception_handler(key, value)

        self.router = Router(
            routes, on_startup=on_startup, on_shutdown=on_shutdown, lifespan=lifespan
        )

    def get(self, path: str, **kwargs: Any) -> Callable[[TCallable], TCallable]:
        """Register GET endpoint to the application."""
        return self.register_route(path, method="GET", **kwargs)

    def head(self, path: str, **kwargs: Any) -> Callable[[TCallable], TCallable]:
        """Register GET endpoint to the application."""
        return self.register_route(path, method="HEAD", **kwargs)

    def post(self, path: str, **kwargs: Any) -> Callable[[TCallable], TCallable]:
        """Register POST endpoint to the application."""
        return self.register_route(path, method="POST", **kwargs)

    def put(self, path: str, **kwargs: Any) -> Callable[[TCallable], TCallable]:
        """Register PUT endpoint to the application."""
        return self.register_route(path, method="PUT", **kwargs)

    def delete(self, path: str, **kwargs: Any) -> Callable[[TCallable], TCallable]:
        """Register DELETE endpoint to the application."""
        return self.register_route(path, method="DELETE", **kwargs)

    def patch(self, path: str, **kwargs: Any) -> Callable[[TCallable], TCallable]:
        """Register PATCH endpoint to the application."""
        return self.register_route(path, method="PATCH", **kwargs)

    def options(self, path: str, **kwargs: Any) -> Callable[[TCallable], TCallable]:
        """Register OPTIONS endpoint to the application."""
        return self.register_route(path, method="OPTIONS", **kwargs)

    def register_route(
        self,
        path: str,
        method: Literal[
            "GET", "HEAD", "POST", "PUT", "DELETE", "PATCH", "OPTIONS"
        ] = "GET",
        name: str | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[TCallable], TCallable]:
        """Register an endpoint to the application."""

        def register(handler: TCallable) -> TCallable:
            self.add_route(
                path,
                handler,
                methods=[method],
                name=name,
                include_in_schema=include_in_schema,
            )
            return handler

        return register

    def endpoint(
        self,
        path: str,
        name: str | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[type[TEndpoint]], type[TEndpoint]]:
        """Register a Ludic class endpoint to the application."""

        def register(endpoint: type[TEndpoint]) -> type[TEndpoint]:
            self.add_route(
                path, endpoint, name=name, include_in_schema=include_in_schema
            )
            return endpoint

        return register

    def add_route(
        self,
        path: str,
        route: Callable[..., Any],
        methods: list[str] | None = None,
        name: str | None = None,
        include_in_schema: bool = True,
    ) -> None:
        """Add a Ludic endpoint to the application.

        The endpoint will be wrapped in a :class:`LudicResponse`
        if a Ludic component is returned. This means that it is possible to return
        a Ludic component instance from the handler directly instead of wrapping it
        in the :class:`LudicResponse`.

        It is still possible to return a regular response from the handler
        (e.g. ``JSONResponse({"hello": "world"})``).
        """
        super().add_route(
            path,
            route,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )

    def url_path_for(self, name: str, /, **path_params: Any) -> URLPath:
        """Get the URL path for an endpoint.

        Args:
            name: The name of the endpoint.
            **path_params: The path parameters.

        Returns:
            The URL path.
        """
        return self.router.url_path_for(name, **path_params)

    def add_exception_handler(
        self,
        exc_class_or_status_code: int | type[Exception],
        handler: TCallable,
    ) -> None:
        """Add an exception handler to the application.

        Example:

            async def not_found(request: Request, exc: HTTPException):
                return Page(
                    Header("Page Could Not Be Found"),
                    Body(f"Here is the reason: {exc.detail}")
                )

            app.add_exception_handler(404, not_found)

        Args:
            exc_class_or_status_code: The exception class or status code.
            handler: The exception handler function.
        """

        @wraps(handler)
        async def wrapped_handler(
            request: Request | WebSocket, exc: Exception
        ) -> Response:
            parameters = inspect.signature(handler).parameters
            handler_kw: dict[str, Any] = {}
            is_async = is_async_callable(handler)

            for name, param in parameters.items():
                if issubclass(param.annotation, Exception):
                    handler_kw[name] = exc
                elif issubclass(param.annotation, Request):
                    handler_kw[name] = request

            if is_async:
                with BaseElement.formatter:
                    response = await handler(**handler_kw)
            else:
                response = await run_in_threadpool_safe(handler, **handler_kw)

            if isinstance(response, BaseElement):
                return LudicResponse(response, getattr(exc, "status_code", 500))
            return cast(Response, response)

        self.exception_handlers[exc_class_or_status_code] = wrapped_handler

    def exception_handler(
        self, exc_class_or_status_code: int | type[Exception]
    ) -> Callable[[TCallable], TCallable]:
        """Register an exception handler to the application.

        Example:

            @app.exception_handler(404)
            async def not_found(request: Request, exc: HTTPException):
                return Page(
                    Header("Page Could Not Be Found"),
                    Body(f"Here is the reason: {exc.detail}")
                )
        """

        def decorator(handler: TCallable) -> TCallable:
            self.add_exception_handler(exc_class_or_status_code, handler)
            return handler

        return decorator



from typing import Annotated, NotRequired, Self, override

from examples import Page, init_db

from ludic.catalog.buttons import Button, ButtonDanger, ButtonPrimary
from ludic.catalog.forms import FieldMeta, Form, create_fields
from ludic.catalog.headers import H1, H2
from ludic.catalog.items import Pairs
from ludic.catalog.layouts import Box, Cluster, Stack
from ludic.catalog.quotes import Quote
from ludic.types import Attrs
from ludic.web import Endpoint, LudicApp
from ludic.web.exceptions import NotFoundError
from ludic.web.parsers import Parser, ValidationError

db = init_db()
app = LudicApp(debug=True)


def email_validator(email: str) -> str:
    if len(email.split("@")) != 2:
        raise ValidationError("Invalid email")
    return email


class ContactAttrs(Attrs):
    id: NotRequired[str]
    first_name: Annotated[str, FieldMeta(label="First Name")]
    last_name: Annotated[str, FieldMeta(label="Last Name")]
    email: Annotated[
        str, FieldMeta(label="Email", type="email", parser=email_validator)
    ]


@app.get("/")
async def index() -> Page:
    return Page(
        H1("Click To Edit"),
        Quote(
            "The click to edit pattern provides a way to offer inline editing "
            "of all or part of a record without a page refresh.",
            source_url="https://htmx.org/examples/click-to-edit/",
        ),
        H2("Demo"),
        Box(*[Contact(**contact.dict()) for contact in db.contacts.values()]),
    )


@app.endpoint("/contacts/{id}")
class Contact(Endpoint[ContactAttrs]):
    @classmethod
    async def get(cls, id: str) -> Self:
        contact = db.contacts.get(id)

        if contact is None:
            raise NotFoundError("Contact not found")

        return cls(**contact.dict())

    @classmethod
    async def put(cls, id: str, attrs: Parser[ContactAttrs]) -> Self:
        contact = db.contacts.get(id)

        if contact is None:
            raise NotFoundError("Contact not found")

        for key, value in attrs.validate().items():
            setattr(contact, key, value)

        return cls(**contact.dict())

    @override
    def render(self) -> Stack:
        return Stack(
            Pairs(items=self.attrs.items()),
            Cluster(
                Button(
                    "Click To Edit",
                    hx_get=self.url_for(ContactForm),
                ),
            ),
            hx_target="this",
            hx_swap="outerHTML",
        )


@app.endpoint("/contacts/{id}/form/")
class ContactForm(Endpoint[ContactAttrs]):
    @classmethod
    async def get(cls, id: str) -> Self:
        contact = db.contacts.get(id)

        if contact is None:
            raise NotFoundError("Contact not found")

        return cls(**contact.dict())

    @override
    def render(self) -> Form:
        return Form(
            *create_fields(self.attrs, spec=ContactAttrs),
            Cluster(
                ButtonPrimary("Submit"),
                ButtonDanger("Cancel", hx_get=self.url_for(Contact)),
            ),
            hx_put=self.url_for(Contact),
            hx_target="this",
            hx_swap="outerHTML",
        )



from typing import override

from ludic.attrs import GlobalAttrs
from ludic.html import li, ol, ul
from ludic.types import AnyChildren, Component


class ListAttrs(GlobalAttrs, total=False):
    items: list[AnyChildren]


class Item(Component[AnyChildren, GlobalAttrs]):
    """Simple component simulating an item in a list.

    Example usage:

        Item("Item 1")
    """

    @override
    def render(self) -> li:
        return li(*self.children, **self.attrs)


class List(Component[Item, ListAttrs]):
    """Simple component simulating a list.

    There is basically just an alias for the :class:`ul` element
    without the requirement to pass `li` as children.

    Example usage:

        List(Item("Item 1"), Item("Item 2"))
    """

    @override
    def render(self) -> ul:
        if items := self.attrs.get("items"):
            children = tuple(map(Item, items))
        else:
            children = self.children
        return ul(*children, **self.attrs)


class NumberedList(Component[Item, ListAttrs]):
    """Simple component simulating a numbered list.

    There is basically just an alias for the :class:`ol` element
    without the requirement to pass `li` as children.

    Example usage:

        NumberedList(Item("Item 1"), Item("Item 2"))
    """

    @override
    def render(self) -> ol:
        if items := self.attrs.get("items"):
            children = tuple(map(Item, items))
        else:
            children = self.children
        return ol(*children, **self.attrs)



import json
from collections.abc import MutableMapping
from typing import Any

from starlette.datastructures import (
    URL,
    FormData,
    QueryParams,
    URLPath,
)
from starlette.datastructures import Headers as BaseHeaders

from ludic import types

__all__ = ("FormData", "Headers", "QueryParams", "URL", "URLPath")


class Headers(BaseHeaders):
    """An immutable, case-insensitive multi-dict representing HTTP headers."""

    def __init__(
        self,
        headers: types.Headers | None = None,
        raw: list[tuple[bytes, bytes]] | None = None,
        scope: MutableMapping[str, Any] | None = None,
    ) -> None:
        new_headers: dict[str, str] = {}

        if headers:
            for key, value in headers.items():
                if isinstance(value, str):
                    new_headers[key] = value
                else:
                    new_headers[key] = json.dumps(value)

        super().__init__(new_headers, raw, scope)



from ludic.html import div, script
from ludic.types import JavaScript, Safe


def test_safe() -> None:
    assert (
        div("Hello <b>World!</b>").to_html()
        == "<div>Hello &lt;b&gt;World!&lt;/b&gt;</div>"
    )
    assert (
        div(Safe("Hello <b>World!</b>")).to_html() == "<div>Hello <b>World!</b></div>"
    )


def test_javascript() -> None:
    assert (
        div("document.write('<h2>HTML</h2>');").to_html()
        == "<div>document.write('&lt;h2&gt;HTML&lt;/h2&gt;');</div>"
    )
    assert (
        script(JavaScript("document.write('<h2>HTML</h2>');")).to_html()
        == "<script>document.write('<h2>HTML</h2>');</script>"
    )



from ludic.web.datastructures import Headers


def test_headers() -> None:
    headers = Headers({"foo": "bar", "bar": {"foo": "baz"}})

    assert headers["foo"] == "bar"
    assert headers["bar"] == '{"foo": "baz"}'



"""An experimental module for creating HTML tables."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Literal, cast, get_type_hints, override

from typing_extensions import TypeVar

from ludic.attrs import GlobalAttrs
from ludic.html import style, table, tbody, td, th, thead, tr
from ludic.types import (
    AnyChildren,
    BaseElement,
    Component,
    ComponentStrict,
    PrimitiveChildren,
    TAttrs,
)
from ludic.utils import get_annotations_metadata_of_type

from .forms import FieldMeta
from .utils import attr_to_camel


@dataclass
class ColumnMeta:
    """Class to be used as an annotation for attributes.

    Example:

        class PersonAttrs(Attrs):
            id: str
            name: Annotated[
                str,
                ColumnMeta(label="Full Name"),
            ]
            email: Annotated[
                str,
                ColumnMeta(label="Email"),
            ]
    """

    identifier: bool = False
    label: str | None = None
    kind: Literal["text"] | FieldMeta = "text"
    parser: Callable[[Any], PrimitiveChildren] | None = None

    def format(self, key: str, value: Any) -> Any:
        if isinstance(self.kind, FieldMeta):
            return self.kind.format(key, value)
        return value

    def parse(self, value: Any) -> PrimitiveChildren:
        if self.kind == "text":
            return value if self.parser is None else self.parser(value)
        return self.kind(value)

    def __call__(self, value: Any) -> PrimitiveChildren:
        return self.parse(value)


class TableRow(Component[AnyChildren, GlobalAttrs]):
    """Simple component rendering as the HTML ``tr`` element."""

    def get_value(self, index: int) -> PrimitiveChildren | None:
        if len(self.children) > index:
            child = self.children[index]
            return child.text if isinstance(child, BaseElement) else child
        return None

    @override
    def render(self) -> tr:
        return tr(
            *(child if isinstance(child, td) else td(child) for child in self.children),
            **self.attrs,
        )


class TableHead(Component[AnyChildren, GlobalAttrs]):
    """Simple component rendering as the HTML ``tr`` element."""

    @property
    def header(self) -> tuple[PrimitiveChildren, ...]:
        return tuple(
            child.text if isinstance(child, BaseElement) else str(child)
            for child in self.children
            if self.children
        )

    @override
    def render(self) -> tr:
        return tr(
            *(child if isinstance(child, th) else th(child) for child in self.children),
            **self.attrs,
        )


THead = TypeVar("THead", bound=BaseElement, default=TableHead)
TRow = TypeVar("TRow", bound=BaseElement, default=TableRow)


class TableAttrs(GlobalAttrs):
    head_attrs: GlobalAttrs
    body_attrs: GlobalAttrs


class Table(ComponentStrict[THead, *tuple[TRow, ...], TableAttrs]):
    """A component rendering as the HTML ``table`` element.

    The component allows :class:`TableHead` and :class:`TableRow` types by default.

    Example:

        Table(
            TableHead("Name", "Age"),
            TableRow("John", 42),
            TableRow("Jane", 23),
        )

    You can also specify different types of header and body:

        Table[PersonHead, PersonRow](
            PersonHead("Name", "Age"),
            PersonRow("John", 42),
        )
    """

    classes = ["table"]
    styles = style.use(
        lambda theme: {
            "table.table": {
                "inline-size": "100%",  # type: ignore
                "border-collapse": "collapse",  # type: ignore
                "thead": {
                    "background-color": theme.colors.light,
                },
                "tr th": {
                    "border": (
                        f"{theme.borders.thin} solid {theme.colors.light.darken(1)}"
                    ),
                    "padding": f"{theme.sizes.xxs} {theme.sizes.xxxs}",
                },
                "tr td": {
                    "border": (
                        f"{theme.borders.thin} solid {theme.colors.light.darken(1)}"
                    ),
                    "padding": f"{theme.sizes.xxs} {theme.sizes.xxxs}",
                },
            },
        }
    )

    @property
    def header(self) -> tuple[PrimitiveChildren, ...]:
        if isinstance(self.children[0], TableHead):
            return self.children[0].header
        return ()

    def getlist(self, key: str) -> list[PrimitiveChildren | None]:
        result: list[PrimitiveChildren | None] = []

        for idx, head in enumerate(self.header):
            if key != head:
                continue

            rows: list[TableRow] = cast(list[TableRow], self.children[1:])
            for row in rows:
                if value := row.get_value(idx):
                    result.append(value)

        return result

    @override
    def render(self) -> table:
        return table(
            thead(self.children[0], **self.attrs.get("head_attrs", {})),
            tbody(*self.children[1:], **self.attrs.get("body_attrs", {})),
            **self.attrs_for(table),
        )


def create_rows(
    attrs_list: list[TAttrs], spec: type[TAttrs], include_id_column: bool = True
) -> tuple[TableHead, *tuple[TableRow, ...]]:
    """Create table rows from the given attributes.

    Example:

        class PersonAttrs(Attrs):
            id: Annotated[int, ColumnMeta(identifier=True)]
            name: Annotated[str, ColumnMeta(label="Full Name")]
            email: Annotated[str, ColumnMeta(label="Email")]

        people = [
            {"id": 1, "name": "John Doe", "email": "john@j.com"},
            {"id": 2, "name": "Jane Smith", "email": "jane@s.com"},
        ]
        rows = create_rows(people, spec=PersonAttrs)

        table = Table(*rows)

    Args:
        attrs_list (list[TAttrs]): The list of attributes to create table rows from.
        spec (type[TAttrs]): The specification of the attributes.

    Returns:
        tuple[TableHead, *tuple[TableRow, ...]]: list of table rows including header.
    """
    annotations = get_type_hints(spec, include_extras=True)
    matadata = get_annotations_metadata_of_type(annotations, ColumnMeta)

    id_col_name: str = "_index"
    headers = []
    for key, col_meta in matadata.items():
        if col_meta.identifier:
            id_col_name = key
        if not col_meta.identifier or include_id_column:
            headers.append(col_meta.label or attr_to_camel(key))

    rows: list[TableRow] = []
    for idx, attrs in enumerate(attrs_list):
        cells: list[AnyChildren] = []

        for key, value in attrs.items():
            if meta := matadata.get(key):
                if meta.identifier and not include_id_column:
                    continue
                name = f"{key}:{id_col_name}:{attrs.get(id_col_name, str(idx))}"
                cells.append(meta.format(name, value))

        if cells:
            rows.append(TableRow(*cells))

    return TableHead(*headers), *rows



import json
from typing import override

from ludic.attrs import Attrs, NoAttrs
from ludic.html import body, head, html, link, meta, script, style, title
from ludic.types import AnyChildren, BaseElement, Component, ComponentStrict


class HtmlHeadAttrs(Attrs, total=False):
    title: str
    favicon: str
    load_styles: bool
    htmx_config: dict[str, str]


class HtmlBodyAttrs(Attrs, total=False):
    htmx_path: str
    htmx_enabled: bool
    htmx_version: str


class Head(Component[AnyChildren, HtmlHeadAttrs]):
    @override
    def render(self) -> head:
        elements: list[BaseElement] = [title(self.attrs.get("title", "Ludic App"))]

        if favicon := self.attrs.get("favicon"):
            elements.append(link(rel="icon", href=favicon, type="image/x-icon"))
        if config := self.attrs.get("htmx_config", {"defaultSwapStyle": "outerHTML"}):
            elements.append(meta(name="htmx-config", content=json.dumps(config)))
        if self.attrs.get("load_styles", True):
            elements.append(style.load(cache=True))

        return head(*elements, *self.children)


class Body(Component[AnyChildren, HtmlBodyAttrs]):
    @override
    def render(self) -> body:
        scripts = []
        if htmx_path := self.attrs.get("htmx_path"):
            scripts.append(script(src=htmx_path))
        elif self.attrs.get("htmx_enabled", "htmx_version" in self.attrs):
            htmx_version = self.attrs.get("htmx_version", "latest")
            scripts.append(script(src=f"https://unpkg.com/htmx.org@{htmx_version}"))
        return body(*self.children, *scripts)


class HtmlPage(ComponentStrict[Head, Body, NoAttrs]):
    styles = style.use(
        lambda theme: {
            # global styling
            "*": {
                "box-sizing": "border-box",
                "max-inline-size": theme.measure,
                "font-family": theme.fonts.plain,
                "font-size": theme.fonts.size,
                "color": theme.colors.dark,
                "overflow-wrap": "break-word",
                "margin": "0",
                "padding": "0",
                "line-height": theme.line_height,
            },
            ("html", "body", "div", "header", "nav", "main", "footer"): {
                "max-inline-size": "none",
            },
            # elements styling
            "h1": {
                "font-size": theme.headers.h1.size,
                "font-family": theme.fonts.serif,
            },
            "h2": {
                "font-size": theme.headers.h2.size,
                "font-family": theme.fonts.serif,
            },
            "h3": {
                "font-size": theme.headers.h3.size,
                "font-family": theme.fonts.serif,
            },
            "h4": {
                "font-size": theme.headers.h4.size,
                "font-family": theme.fonts.serif,
            },
            "h5": {
                "font-size": theme.headers.h5.size,
                "font-family": theme.fonts.serif,
            },
            "h6": {
                "font-size": theme.headers.h6.size,
                "font-family": theme.fonts.serif,
            },
            "a": {
                "color": theme.colors.primary.darken(2),
                "text-decoration": "none",
            },
            "a:hover": {
                "text-decoration": "underline",
            },
            "pre": {
                "overflow": "auto",
            },
            ("code", "pre", "pre *"): {
                "font-family": theme.fonts.mono,
            },
            "dl": {
                "margin-block": "0",
            },
            "dl dd + dt": {
                "margin-block-start": theme.sizes.xs,
            },
            "dl dt + dd": {
                "margin-block-start": theme.sizes.xxxxs,
            },
            "dt": {
                "font-weight": "bold",
            },
            "dd": {
                "margin-left": "0",
            },
            ("ul", "ol"): {
                "padding-inline-start": theme.sizes.xxl,
            },
            ("ul > li + li", "ol > li + li", "li > * + *"): {
                "margin-block-start": theme.sizes.xs,
            },
            "ul > li::marker": {
                "font-size": theme.fonts.size * 1.2,
            },
            ("img", "svg"): {
                "width": "100%",
            },
            # utilities
            ".text-align-center": {
                "text-align": "center !important",  # type: ignore[typeddict-item]
            },
            ".text-align-right": {
                "text-align": "right !important",  # type: ignore[typeddict-item]
            },
            ".text-align-left": {
                "text-align": "left !important",  # type: ignore[typeddict-item]
            },
            ".justify-space-between": {
                "justify-content": "space-between !important",  # type: ignore[typeddict-item]
            },
            ".justify-space-around": {
                "justify-content": "space-around !important",  # type: ignore[typeddict-item]
            },
            ".justify-space-evenly": {
                "justify-content": "space-evenly !important",  # type: ignore[typeddict-item]
            },
            ".justify-center": {
                "justify-content": "center !important",  # type: ignore[typeddict-item]
            },
            ".justify-end": {
                "justify-content": "end !important",  # type: ignore[typeddict-item]
            },
            ".justify-start": {
                "justify-content": "start !important",  # type: ignore[typeddict-item]
            },
            ".align-center": {
                "align-items": "center !important",  # type: ignore[typeddict-item]
            },
            ".align-end": {
                "align-items": "end !important",  # type: ignore[typeddict-item]
            },
            ".align-start": {
                "align-items": "start !important",  # type: ignore[typeddict-item]
            },
            ".no-padding": {
                "padding": "0 !important",
            },
            ".no-inline-padding": {
                "padding-inline": "0 !important",
            },
            ".no-margin": {
                "margin": "0 !important",
            },
            ".no-inline-margin": {
                "margin-inline": "0 !important",
            },
            ".no-block-margin": {
                "margin-block": "0 !important",
            },
            ".no-block-padding": {
                "padding-block": "0 !important",
            },
            ".no-border": {
                "border": "none !important",
                "border-radius": "0 !important",
            },
            ".no-border-radius": {
                "border-radius": "0 !important",
            },
        }
    )

    @override
    def render(self) -> html:
        return html(
            self.children[0].render(),
            self.children[1].render(),
        )



import inspect
from collections.abc import Callable
from typing import Any, ClassVar, Protocol

from starlette.datastructures import URL
from starlette.endpoints import HTTPEndpoint as BaseEndpoint
from starlette.routing import Route

from ludic.catalog.loaders import LazyLoader
from ludic.types import AnyChildren, Component, NoChildren, TAttrs
from ludic.utils import get_element_generic_args

from .requests import Request
from .responses import prepare_response


class RoutedProtocol(Protocol):
    route: ClassVar[Route]


class HTTPEndpoint(BaseEndpoint):
    """Endpoint class for Ludic components.

    Usage:

        class MyButtonEndpoint(HTTPEndpoint):
            async def get(self, request: Request) -> button:
                return button(...)
    """

    route: ClassVar[Route]

    async def dispatch(self) -> None:
        request = Request(self.scope, receive=self.receive)
        handler_name = (
            "get"
            if request.method == "HEAD" and not hasattr(self, "head")
            else request.method.lower()
        )

        handler: Callable[..., Any] = (
            getattr(self, handler_name, None) or self.method_not_allowed
        )
        response = await prepare_response(handler, request)

        await response(self.scope, self.receive, self.send)


class Endpoint(Component[NoChildren, TAttrs]):
    """Base class for Ludic endpoints."""

    route: ClassVar[Route]

    @property
    def request(self) -> Request | None:
        """The current request."""
        return self.context.get("request")

    def lazy_load(
        self,
        endpoint: type[RoutedProtocol],
        placeholder: AnyChildren = "Loading ...",
        **kwargs: Any,
    ) -> LazyLoader:
        """Lazy load an endpoint's content.

        Args:
            endpoint (type[RoutedProtocol]): The endpoint to lazy load.
            placeholder (AnyChild): The content to show while loading.
            **kwargs: URL path parameters.
        """
        return LazyLoader(
            placeholder=placeholder,
            load_url=self.url_for(endpoint, **kwargs),
        )

    def url_for(self, endpoint: type[RoutedProtocol] | str, **path_params: Any) -> URL:
        """Get URL for an endpoint.

        Args:
            endpoint: The endpoint.
            **path_params: URL path parameters.

        Returns:
            The URL.
        """
        if self.request is None or not isinstance(self.request, Request):
            raise RuntimeError(
                f"{type(self).__name__} is not bound to a request, you can only use "
                f"the {type(self).__name__}.url_for method in the context of a request."
            )

        if inspect.isclass(endpoint) and issubclass(endpoint, Endpoint):
            endpoint_generic_args = get_element_generic_args(endpoint)
            self_generic_args = get_element_generic_args(self)

            if (
                endpoint_generic_args
                and self_generic_args
                and endpoint_generic_args[-1] is self_generic_args[-1]
            ):
                path_params = {
                    key: value
                    for key, value in {**self.attrs, **path_params}.items()
                    if key in endpoint.route.param_convertors
                }

        return self.request.url_for(endpoint, **path_params)



from typing import Self, override

from examples import Page, init_db

from ludic.attrs import Attrs, HtmxAttrs
from ludic.catalog.buttons import ButtonDanger
from ludic.catalog.headers import H1, H2
from ludic.catalog.quotes import Quote
from ludic.catalog.tables import Table, TableHead, TableRow
from ludic.web import Endpoint, LudicApp
from ludic.web.exceptions import NotFoundError

db = init_db()
app = LudicApp(debug=True)


class PersonAttrs(Attrs):
    id: str
    name: str
    email: str
    active: bool


class PeopleAttrs(Attrs):
    people: list[PersonAttrs]


@app.get("/")
def index() -> Page:
    return Page(
        H1("Delete Row"),
        Quote(
            "This example shows how to implement a delete button that removes "
            "a table row upon completion.",
            source_url="https://htmx.org/examples/delete-row/",
        ),
        H2("Demo"),
        PeopleTable.get(),
    )


@app.endpoint("/people/{id}")
class PersonRow(Endpoint[PersonAttrs]):
    @classmethod
    def delete(cls, id: str) -> None:
        try:
            db.people.pop(id)
        except KeyError:
            raise NotFoundError("Person not found")

    @override
    def render(self) -> TableRow:
        return TableRow(
            self.attrs["name"],
            self.attrs["email"],
            "Active" if self.attrs["active"] else "Inactive",
            ButtonDanger(
                "Delete", hx_delete=self.url_for(PersonRow), classes=["small"]
            ),
        )


@app.endpoint("/people/")
class PeopleTable(Endpoint[PeopleAttrs]):
    styles = {
        "tr.htmx-swapping td": {
            "opacity": "0",
            "transition": "opacity 1s ease-out",
        }
    }

    @classmethod
    def get(cls) -> Self:
        return cls(people=[person.dict() for person in db.people.values()])

    @override
    def render(self) -> Table[TableHead, PersonRow]:
        return Table(
            TableHead("Name", "Email", "Active", ""),
            *(PersonRow(**person) for person in self.attrs["people"]),
            body_attrs=HtmxAttrs(
                hx_confirm="Are you sure?",
                hx_target="closest tr",
                hx_swap="outerHTML swap:1s",
            ),
            classes=["text-align-center"],
        )



from starlette.testclient import TestClient


def test_bulk_update() -> None:
    from examples.bulk_update import app, db

    with TestClient(app) as client:
        assert client.get("/").status_code == 200
        assert client.get("/people/").status_code == 200
        assert db.people["1"].active
        assert db.people["2"].active
        assert db.people["3"].active
        assert not db.people["4"].active

        activate_data = {"active:id:1": "on", "active:id:2": "on"}
        assert client.post("/people/", data=activate_data).status_code == 200
        assert db.people["1"].active
        assert db.people["2"].active
        assert not db.people["3"].active
        assert not db.people["4"].active


def test_click_to_edit() -> None:
    from examples.click_to_edit import app, db

    with TestClient(app) as client:
        assert client.get("/").status_code == 200
        assert client.get("/contacts/1").status_code == 200
        assert client.get("/contacts/1/form/").status_code == 200
        assert client.get("/contacts/123").status_code == 404
        assert client.get("/contacts/123/form/").status_code == 404
        assert db.contacts["1"].first_name == "John"

        edit_data = {
            "first_name": "Test",
            "last_name": "Doe",
            "email": "test@example.com",
        }
        assert client.put("/contacts/1", data=edit_data).status_code == 200
        assert db.contacts["1"].first_name == "Test"


def test_click_to_load() -> None:
    from examples.click_to_load import app

    with TestClient(app) as client:
        assert client.get("/").status_code == 200
        assert client.get("/contacts/").status_code == 200
        assert client.get("/contacts/?page=2").status_code == 200


def test_delete_row() -> None:
    from examples.delete_row import app, db

    with TestClient(app) as client:
        assert client.get("/").status_code == 200
        assert client.get("/people/").status_code == 200
        assert client.delete("/people/1").status_code == 200
        assert client.delete("/people/123").status_code == 404
        assert db.people.get("1") is None


def test_edit_row() -> None:
    from examples.edit_row import app, db

    with TestClient(app) as client:
        assert client.get("/").status_code == 200
        assert client.get("/people/").status_code == 200
        assert client.get("/people/1").status_code == 200
        assert client.get("/people/1/form/").status_code == 200
        assert client.get("/people/123").status_code == 404
        assert client.get("/people/123/form/").status_code == 404

        assert db.people["1"].name == "Joe Smith"
        edit_data = {"name": "Test", "email": "test@example.com"}
        assert client.put("/people/1", data=edit_data).status_code == 200
        assert db.people["1"].name == "Test"
        assert db.people["1"].email == "test@example.com"


def test_lazy_loading() -> None:
    from examples.lazy_loading import app

    with TestClient(app) as client:
        assert client.get("/").status_code == 200
        response = client.get("/load/0")
        assert response.status_code == 200
        assert b"Content Loaded" in response.content


def test_infinite_scroll() -> None:
    from examples.infinite_scroll import app

    with TestClient(app) as client:
        assert client.get("/").status_code == 200
        assert client.get("/contacts/").status_code == 200
        assert client.get("/contacts/?page=2").status_code == 200



from typing import override

from ludic.attrs import GlobalAttrs
from ludic.html import a, b, div, style
from ludic.styles.themes import (
    Colors,
    Fonts,
    get_default_theme,
    set_default_theme,
)
from ludic.styles.types import Color, Size
from ludic.types import Component

from . import BarTheme, FooTheme


def test_theme_colors() -> None:
    theme = FooTheme(
        colors=Colors(
            primary=Color("#c2e7fd"),
            white=Color("#fff"),
            light=Color("#eee"),
            dark=Color("#333"),
            black=Color("#000"),
        )
    )

    assert theme.colors.primary.rgb == (194, 231, 253)
    assert theme.colors.white.rgb == (255, 255, 255)
    assert theme.colors.light.rgb == (238, 238, 238)
    assert theme.colors.dark.rgb == (51, 51, 51)
    assert theme.colors.black.rgb == (0, 0, 0)

    assert theme.colors.white.darken(10).rgb == (127, 127, 127)
    assert theme.colors.white.darken(20).rgb == (1, 1, 1)
    assert theme.colors.black.lighten(10).rgb == (127, 127, 127)
    assert theme.colors.black.lighten(20).rgb == (255, 255, 255)

    assert theme.colors.light.darken(10).rgb == (119, 119, 119)
    assert theme.colors.dark.lighten(10).rgb == (153, 153, 153)

    assert theme.colors.primary.darken(10).rgb == (97, 115, 126)


def test_theme_font_sizes() -> None:
    theme = FooTheme(fonts=Fonts(size=Size(10, "px")))

    assert theme.fonts.size == "10px"
    assert theme.fonts.plain == "Helvetica Neue, Helvetica, Arial, sans-serif"

    assert theme.fonts.size - 1 == "9px"
    assert theme.fonts.size + 5 == "15px"
    assert theme.fonts.size * 2 == "20px"

    assert theme.fonts.size + "2" == "10px"
    assert theme.fonts.size - "ab" == "10px"


def test_themes_switching() -> None:
    foo, bar = FooTheme(), BarTheme()

    set_default_theme(foo)
    assert get_default_theme() == foo
    set_default_theme(bar)
    assert get_default_theme() == bar


def test_element_theme_switching() -> None:
    foo = FooTheme()
    bar = BarTheme()

    set_default_theme(bar)

    class C1(Component[str, GlobalAttrs]):
        styles = style.use(
            lambda theme: {
                "#c1 a": {"color": theme.colors.warning},
            }
        )

        @override
        def render(self) -> div:
            return div(
                b("Hello, ", style={"color": self.theme.colors.secondary}),
                a(*self.children, href="https://example.com"),
                id="c1",
            )

    class C2(Component[str, GlobalAttrs]):
        styles = style.use(
            lambda theme: {
                "#c2 a": {"color": theme.colors.danger},
            }
        )

        @override
        def render(self) -> div:
            return div(
                foo.use(C1(*self.children)),
                id="c2",
                style={"background-color": self.theme.colors.primary},
            )

    assert C2("World").to_html() == (
        f'<div id="c2" style="background-color:{bar.colors.primary}">'
          '<div id="c1">'
            f'<b style="color:{foo.colors.secondary}">Hello, </b>'
            '<a href="https://example.com">World</a>'
          "</div>"
        "</div>"
    )  # fmt: skip
    assert style.from_components(C1, C2).to_html() == (
        '<style type="text/css">\n'
          f"#c1 a {{ color: {bar.colors.warning}; }}\n"
          f"#c2 a {{ color: {bar.colors.danger}; }}\n"
        "</style>"
    )  # fmt: skip

    set_default_theme(foo)

    assert style.from_components(C1, C2).to_html() == (
        '<style type="text/css">\n'
          f"#c1 a {{ color: {foo.colors.warning}; }}\n"
          f"#c2 a {{ color: {foo.colors.danger}; }}\n"
        "</style>"
    )  # fmt: skip



import inspect
from collections.abc import Callable
from typing import Any

from starlette import routing
from starlette.exceptions import HTTPException
from starlette.responses import PlainTextResponse, Response
from starlette.routing import Host
from starlette.types import Receive, Scope, Send

from ludic.attrs import Attrs

from .endpoints import Endpoint
from .requests import Request
from .responses import prepare_response

__all__ = (
    "Host",
    "Mount",
    "Route",
    "Router",
)


class Mount(routing.Mount):
    """Mount class for Ludic components."""

    def matches(self, scope: Scope) -> tuple[routing.Match, Scope]:
        match, scope = super().matches(scope)
        if match == routing.Match.FULL:
            if partial := scope.get("partial_mount"):
                scope["partial_mount"] = f"{partial}:{self.name}"
            else:
                scope["partial_mount"] = self.name
        return match, scope


class _FunctionHandler:
    def __init__(self, handler: Callable[..., Any]) -> None:
        self.handler = handler

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive)
        response = await prepare_response(self.handler, request)
        await response(scope, receive, send)


class _EndpointHandler:
    def __init__(self, handler: type[Endpoint[Attrs]]) -> None:
        self.handler = handler
        self._allowed_methods = [
            method
            for method in ("GET", "HEAD", "POST", "PUT", "PATCH", "DELETE", "OPTIONS")
            if getattr(self.handler, method.lower(), None) is not None
        ]

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive)
        handler_name = (
            "get"
            if request.method == "HEAD" and not hasattr(self, "head")
            else request.method.lower()
        )
        handler: Callable[..., Any] = getattr(
            self.handler, handler_name, self.method_not_allowed(scope)
        )
        response = await prepare_response(handler, request)
        await response(scope, receive, send)

    def method_not_allowed(self, scope: Scope) -> Callable[..., Any]:
        # If we're running inside a starlette application then raise an
        # exception, so that the configurable exception handler can deal with
        # returning the response. For plain ASGI apps, just return the response.
        headers = {"Allow": ", ".join(self._allowed_methods)}

        async def make_response(request: Request) -> Response:
            if "app" in scope:
                raise HTTPException(status_code=405, headers=headers)
            return PlainTextResponse(
                "Method Not Allowed", status_code=405, headers=headers
            )

        return make_response


class Route(routing.Route):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        name = routing.get_name(endpoint) if name is None else name
        wrapped_route = endpoint
        if inspect.isfunction(endpoint) or inspect.ismethod(endpoint):
            wrapped_route = _FunctionHandler(endpoint)
        elif inspect.isclass(endpoint) and issubclass(endpoint, Endpoint):
            wrapped_route = _EndpointHandler(endpoint)
        if getattr(endpoint, "route", None) is None:
            endpoint.route = self  # type: ignore
        super().__init__(path, wrapped_route, name=name, **kwargs)


class Router(routing.Router):
    def add_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        methods: list[str] | None = None,
        name: str | None = None,
        include_in_schema: bool = True,
    ) -> None:
        route = Route(
            path,
            endpoint=endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )
        self.routes.append(route)



from typing import Self, override

from examples import Page

from ludic.catalog.buttons import ButtonPrimary
from ludic.catalog.headers import H1, H2
from ludic.catalog.quotes import Quote
from ludic.catalog.tables import Table, TableHead, TableRow
from ludic.html import td
from ludic.types import Attrs, Blank, Component, ComponentStrict
from ludic.web import Endpoint, LudicApp
from ludic.web.datastructures import QueryParams

app = LudicApp(debug=True)


class ContactAttrs(Attrs):
    id: str
    name: str
    email: str


class ContactsSliceAttrs(Attrs):
    page: int
    contacts: list[ContactAttrs]


class LoadMoreAttrs(Attrs):
    url: str


def load_contacts(page: int) -> list[ContactAttrs]:
    return [
        ContactAttrs(
            id=str(page * 10 + idx),
            email=f"void{page * 10 + idx}@null.org",
            name="Agent Smith",
        )
        for idx in range(10)
    ]


class LoadMoreButton(ComponentStrict[LoadMoreAttrs]):
    target: str = "replace-me"

    @override
    def render(self) -> ButtonPrimary:
        return ButtonPrimary(
            "Load More Agents...",
            hx_get=self.attrs["url"],
            hx_target=f"#{self.target}",
            hx_swap="outerHTML",
        )


@app.get("/")
async def index() -> Page:
    return Page(
        H1("Click To Load"),
        Quote(
            "This example shows how to implement click-to-load the next page in "
            "a table of data.",
            source_url="https://htmx.org/examples/click-to-load/",
        ),
        H2("Demo"),
        ContactsTable(await ContactsSlice.get(QueryParams(page=1))),
    )


@app.endpoint("/contacts/")
class ContactsSlice(Endpoint[ContactsSliceAttrs]):
    @classmethod
    async def get(cls, params: QueryParams) -> Self:
        page = int(params.get("page", 1))
        return cls(page=page, contacts=load_contacts(page))

    @override
    def render(self) -> Blank[TableRow]:
        next_page = self.attrs["page"] + 1
        return Blank(
            *(
                TableRow(contact["id"], contact["name"], contact["email"])
                for contact in self.attrs["contacts"]
            ),
            TableRow(
                td(
                    LoadMoreButton(
                        url=self.url_for(ContactsSlice).include_query_params(
                            page=next_page
                        ),
                    ),
                    colspan=3,
                ),
                id=LoadMoreButton.target,
            ),
        )


class ContactsTable(Component[ContactsSlice, Attrs]):
    @override
    def render(self) -> Table[TableHead, ContactsSlice]:
        return Table(
            TableHead("ID", "Name", "Email"),
            *self.children,
            classes=["text-align-center"],
        )



from typing import NotRequired, override

from ludic.attrs import GlobalAttrs
from ludic.html import h2, li, nav, style, ul
from ludic.types import Component, ComponentStrict, PrimitiveChildren

from .buttons import ButtonLink


class NavItemAttrs(GlobalAttrs):
    to: str
    active: NotRequired[bool]
    active_subsection: NotRequired[bool]


class NavHeader(ComponentStrict[str, GlobalAttrs]):
    """Simple component simulating a navigation header.

    This component is supposed to be used as a child of the :class:`Navigation`
    component.
    """

    classes = ["nav-header"]
    styles = style.use(
        lambda theme: {
            "h2.nav-header": {
                "font-size": theme.headers.h2.size * 0.7,
            }
        }
    )

    @override
    def render(self) -> h2:
        return h2(self.children[0], **self.attrs_for(h2))


class NavItem(Component[PrimitiveChildren, NavItemAttrs]):
    """Simple component simulating a navigation item.

    This component is supposed to be used as a child of the :class:`Navigation`
    component.
    """

    classes = ["nav-item"]
    styles = style.use(
        lambda theme: {
            "li.nav-item.section.active-subsection .btn": {
                "font-weight": "bold",
                "background": "none",
            },
            "li.nav-item.subsection": {
                "padding-inline-start": theme.sizes.m,
            },
        }
    )

    @override
    def render(self) -> li:
        self.attrs.setdefault("classes", [])

        if self.attrs.pop("active_subsection", False):
            self.attrs["classes"].append("active-subsection")

        if self.attrs.pop("active", False):
            self.attrs["classes"].append("active")

        return li(
            ButtonLink(
                self.children[0],
                to=self.attrs["to"],
                external=False,
                classes=["small"] if "subsection" in self.attrs["classes"] else [],
            ),
            **self.attrs_for(li),
        )


class NavSection(ComponentStrict[NavHeader, *tuple[NavItem, ...], GlobalAttrs]):
    """Simple component simulating a navigation section.

    This component is supposed to be used as a child of the :class:`Navigation`
    component.
    """

    @override
    def render(self) -> li:
        self.attrs.setdefault("classes", ["stack", "tiny"])
        return li(
            self.children[0],
            ul(*self.children[1:], classes=["stack", "tiny"]),
            **self.attrs_for(li),
        )


class Navigation(Component[NavItem | NavSection, GlobalAttrs]):
    """Simple component simulating a navigation bar.

    Example usage:

        Navigation(
            NavItem("Home", to="/"),
            NavItem("About", to="/about"),
        )
    """

    classes = ["navigation"]
    styles = {
        "nav.navigation ul": {
            "list-style": "none",
            "padding": "0",
        },
    }

    @override
    def render(self) -> nav:
        return nav(
            ul(*self.children, classes=["stack", "small"]),
            **self.attrs_for(nav),
        )



from typing import override

from ludic.attrs import Attrs, GlobalAttrs
from ludic.html import a, div, h1, h2, h3, h4, style
from ludic.types import Component, ComponentStrict

from .utils import text_to_kebab


class AnchorAttrs(Attrs):
    target: str


class Anchor(Component[str, AnchorAttrs]):
    """Component representing a clickable anchor."""

    classes = ["anchor"]
    styles = style.use(
        lambda theme: {
            "a.anchor": {
                "font-family": theme.fonts.serif,
                "color": theme.colors.light.darken(1),
                "text-decoration": "none",
            },
            "a.anchor:hover": {
                "color": theme.colors.dark,
                "text-decoration": "none",
            },
        }
    )

    @override
    def render(self) -> a:
        return a(
            self.children[0] if self.children else "#", href=f"#{self.attrs["target"]}"
        )


class WithAnchorAttrs(GlobalAttrs, total=False):
    anchor: Anchor | bool


class WithAnchor(ComponentStrict[h1 | h2 | h3 | h4 | str, WithAnchorAttrs]):
    """Component which renders its content (header) with a clickable anchor."""

    classes = ["with-anchor"]
    styles = style.use(
        lambda theme: {
            ".with-anchor": {
                "display": "flex",
                "flex-wrap": "wrap",
                "justify-content": "flex-start",
            },
            ".with-anchor > h1 + a": {
                "margin-inline-start": theme.sizes.m,
                "font-size": theme.headers.h1.size,
            },
            ".with-anchor > h2 + a": {
                "margin-inline-start": theme.sizes.s,
                "font-size": theme.headers.h2.size,
            },
            ".with-anchor > h3 + a": {
                "margin-inline-start": theme.sizes.xs,
                "font-size": theme.headers.h3.size,
            },
            ".with-anchor > h4 + a": {
                "margin-inline-start": theme.sizes.xxs,
                "font-size": theme.headers.h4.size,
            },
        }
    )

    @override
    def render(self) -> div:
        element: h1 | h2 | h3 | h4
        if isinstance(self.children[0], h1 | h2 | h3 | h4):
            element = self.children[0]
        else:
            element = h1(*self.children)

        element.attrs.setdefault("id", text_to_kebab(element.text))
        id = element.attrs["id"]

        return div(
            element,
            (
                self.attrs["anchor"]
                if isinstance(self.attrs.get("anchor"), Anchor)
                else Anchor(target=id)
            ),
        )


class H1(ComponentStrict[str, WithAnchorAttrs]):
    """Component rendering as h1 with an optional clickable anchor."""

    @override
    def render(self) -> h1 | WithAnchor:
        header = h1(*self.children, **self.attrs_for(h1))
        anchor = self.attrs.get("anchor")
        if anchor:
            return WithAnchor(header, anchor=anchor)
        elif self.theme.headers.h1.anchor and anchor is not False:
            return WithAnchor(header)
        else:
            return header


class H2(ComponentStrict[str, WithAnchorAttrs]):
    """Component rendering as h2 with an optional clickable anchor."""

    @override
    def render(self) -> h2 | WithAnchor:
        header = h2(*self.children, **self.attrs_for(h2))
        anchor = self.attrs.get("anchor")
        if anchor:
            return WithAnchor(header, anchor=anchor)
        elif self.theme.headers.h2.anchor and anchor is not False:
            return WithAnchor(header)
        else:
            return header


class H3(ComponentStrict[str, WithAnchorAttrs]):
    """Component rendering as h3 with an optional clickable anchor."""

    @override
    def render(self) -> h3 | WithAnchor:
        header = h3(*self.children, **self.attrs_for(h3))
        anchor = self.attrs.get("anchor")
        if anchor:
            return WithAnchor(header, anchor=anchor)
        elif self.theme.headers.h3.anchor and anchor is not False:
            return WithAnchor(header)
        else:
            return header


class H4(ComponentStrict[str, WithAnchorAttrs]):
    """Component rendering as h4 with an optional clickable anchor."""

    @override
    def render(self) -> h4 | WithAnchor:
        header = h4(*self.children, **self.attrs_for(h4))
        anchor = self.attrs.get("anchor")
        if anchor:
            return WithAnchor(header, anchor=anchor)
        elif self.theme.headers.h4.anchor and anchor is not False:
            return WithAnchor(header)
        else:
            return header



import html
import random
import re
from contextvars import ContextVar
from typing import Annotated, Any, Final, TypeVar, get_args, get_origin

from .utils import get_element_attrs_annotations

EXTRACT_NUMBER_RE: Final[re.Pattern[str]] = re.compile(r"\{(\d+:id)\}")

T = TypeVar("T")


def format_attr_value(key: str, value: Any, is_html: bool = False) -> str:
    """Format an HTML attribute with the given key and value.

    Args:
        key (str): The key of the attribute.
        value (Any): The value of the attribute, can be a string or a dictionary.
    Returns:
        str: The formatted HTML attribute.
    """
    if isinstance(value, dict):
        value = ";".join(
            f"{dict_key}:{html.escape(dict_value, False)}"
            for dict_key, dict_value in value.items()
        )
    elif isinstance(value, list):
        value = " ".join(html.escape(v, False) for v in value)
    elif isinstance(value, bool):
        if is_html and not key.startswith("hx"):
            value = html.escape(key, False) if value else ""
        else:
            value = "true" if value else "false"
    elif isinstance(value, str) and getattr(value, "escape", True):
        value = html.escape(value, False)
    return str(value)


def format_attrs(
    attrs_type: Any, attrs: dict[str, Any], is_html: bool = False
) -> dict[str, Any]:
    """Format the given attributes according to the element's attributes.

    Here is an example of TypedDict definition:

        class PersonAttrs(TypedDict):
            name: str
            class_: Annotated[str, "class"]
            is_adult: bool

    And here is the attrs that will be formatted:

        attrs = {"name": "John", "class_": "person", "is_adult": True}

    The result will be:

        >>> format_attrs(PersonAttrs, attrs)
        >>> {"name": "John", "class": "person"}

    Args:
        attrs_type (Any): The element.
        attrs (dict[str, Any]): The attributes to format.

    Returns:
        dict[str, Any]: The formatted attributes.
    """
    hints = get_element_attrs_annotations(attrs_type, include_extras=True)

    def _get_key(key: str) -> str:
        if get_origin(hints[key]) is Annotated:
            args = get_args(hints[key])
            if len(args) > 1 and isinstance(args[1], str):
                return args[1]
        return key

    result: dict[str, str] = {}
    for key, value in attrs.items():
        if formatted_value := format_attr_value(key, value, is_html=is_html):
            alias = _get_key(key)
            if alias in result:
                result[alias] += " " + formatted_value
            else:
                result[alias] = formatted_value
    return result


def format_element(child: Any) -> str:
    """Default HTML formatter.

    Args:
        child (AnyChild): The HTML element or text to format.
    """
    if isinstance(child, str) and getattr(child, "escape", True):
        return html.escape(child, False)
    elif hasattr(child, "to_html"):
        return child.to_html()  # type: ignore
    else:
        return str(child)


def _extract_match(text: str | Any) -> int | str:
    if text.endswith(":id") and text[:-3].isdigit():
        return int(text[:-3])
    return str(text)


def extract_identifiers(text: str) -> list[str | int]:
    """Extract numbers from a string.

    Args:
        text (str): The string to extract numbers from.

    Returns:
        Iterable[int]: The extracted numbers.
    """
    parts = [_extract_match(match) for match in EXTRACT_NUMBER_RE.split(text) if match]
    if any(isinstance(part, int) for part in parts):
        return parts
    else:
        return []


class FormatContext:
    """Format context helper for using f-strings in elements.

    Facilitates dynamic content formatting within Ludic elements using f-strings.

    This class addresses potential memory leaks when using f-strings directly in
    element generation. Employing the 'with' statement ensures proper cleanup of
    formatting context.

    It is possible to use f-strings in elements without the contextmanager,
    however, the contextmanager clears the context (cache) at the end of the block.
    So without this manager, you can have memory leaks in your app.

    It is recommended to wrap, for example, request context with this manager.

    Example usage:

        with FormatContext():
            dom = div(f"test {b('foo')} {i('bar')}")

        assert dom.to_html() == "<div>test <b>foo</b> <i>bar</i></div>"
    """

    _context: ContextVar[dict[int, Any]]

    def __init__(self, name: str) -> None:
        self._context = ContextVar(name)

    def get(self) -> dict[int, Any]:
        try:
            return self._context.get()
        except LookupError:
            return {}

    def append(self, obj: Any) -> str:
        """Store the given object in context memory and return the identifier.

        Args:
            obj (Any): The object to store in context memory.

        Returns:
            str: The identifier of the stored object.
        """
        random_id = random.getrandbits(256)

        try:
            cache = self._context.get()
            cache[random_id] = obj
            self._context.set(cache)
        except LookupError:
            self._context.set({random_id: obj})

        return f"{{{random_id}:id}}"

    def extract(self, *args: Any) -> list[Any]:
        """Extract identifiers from the given arguments.

        Example:

            with FormatContext() as ctx:
                first = ctx.append("foo")
                second = ctx.append({"bar": "baz"})
                print(ctx.extract(f"test {first} {second}"))

            ["test ", "foo", " ", {"bar": "baz"}]

        Args:
            args (Any): The arguments to extract identifiers from.

        Returns:
            Any: The extracted arguments.
        """
        extracted_args: list[Any] = []
        for arg in args:
            if isinstance(arg, str) and (parts := extract_identifiers(arg)):
                cache = self.get()
                extracted_args.extend(
                    cache.pop(part) if isinstance(part, int) else part
                    for part in parts
                    if not isinstance(part, int) or part in cache
                )
                self._context.set(cache)
            else:
                extracted_args.append(arg)
        return extracted_args

    def clear(self) -> None:
        """Clear the context memory."""
        self._context.set({})

    def __enter__(self) -> "FormatContext":
        return self

    def __exit__(self, *_: Any) -> None:
        self.clear()



from ludic.html import style
from ludic.types import AnyChildren, Attrs, Component

from . import FooTheme


class A(Component[AnyChildren, Attrs]):
    styles = {
        "a": {
            "color": "red",
        },
        ".content": {
            "background": "#ffe",
            "padding": "10px",
        },
        ".content > p": {
            "color": "blue",
            "font-size": "20px",
        },
    }


class B(Component[AnyChildren, Attrs]):
    styles = {
        "a.test": {
            "color": "blue",
        },
    }


def test_styles_formatting() -> None:
    assert style(
        {
            "a": {
                "text-decoration": "none",
                "color": "red",
            },
            "a:hover": {
                "color": "blue",
                "text-decoration": "underline",
            },
        }
    ).to_html() == (
        "<style>\n"
        "a { text-decoration: none; color: red; }\n"
        "a:hover { color: blue; text-decoration: underline; }\n"
        "</style>"
    )


def test_styles_nested_formatting() -> None:
    assert (
        style(
            {
                "p.message": {
                    "color": "black",  # type: ignore[dict-item]
                    "background": "yellow",  # type: ignore[dict-item]
                    "padding": "10px",  # type: ignore[dict-item]
                    "a": {
                        "color": "red",
                        "text-decoration": "none",
                    },
                    "a:hover": {
                        "text-decoration": "underline",
                    },
                },
            }
        ).to_html()
        == (
            "<style>\n"
            "p.message { color: black; background: yellow; padding: 10px; }\n"
            "p.message a { color: red; text-decoration: none; }\n"
            "p.message a:hover { text-decoration: underline; }\n"
            "</style>"
        )
    )


def test_styles_with_at_rule() -> None:
    assert (
        style(
            {
                ".htmx-settling": {
                    "opacity": "100",
                },
                "@keyframes lds-ellipsis1": {
                    "0%": {
                        "transform": "scale(0)",
                        "color": "red",
                    }
                },
                "@layer state": {
                    ".alert": {
                        "background-color": "brown",
                    },
                    "p": {
                        "padding": "10px",
                    },
                },
            }
        ).to_html()
    ) == (
        "<style>\n"
        ".htmx-settling { opacity: 100; }\n"
        "@keyframes lds-ellipsis1 { 0% { transform: scale(0); color: red; } }\n"
        "@layer state { .alert { background-color: brown; } p { padding: 10px; } }\n"
        "</style>"
    )


def test_styles_collection() -> None:
    assert style.from_components(A, B, theme=FooTheme()).to_html() == (
        '<style type="text/css">\n'
        "a { color: red; }\n"
        ".content { background: #ffe; padding: 10px; }\n"
        ".content > p { color: blue; font-size: 20px; }\n"
        "a.test { color: blue; }\n"
        "</style>"
    )



from abc import ABCMeta
from collections.abc import Callable, Iterator, Mapping, MutableMapping, Sequence
from typing import (
    Any,
    ClassVar,
    Generic,
    Never,
    TypeAlias,
    TypedDict,
    TypeVar,
    TypeVarTuple,
    Unpack,
    cast,
)

from .format import FormatContext, format_attrs, format_element
from .styles import GlobalStyles, Theme, get_default_theme
from .utils import get_element_attrs_annotations

ELEMENT_REGISTRY: MutableMapping[str, list[type["BaseElement"]]] = {}


class Safe(str):
    """Marker for a string that is safe to use as is without HTML escaping.

    That means the content of the string is not escaped when rendered.

    Usage:

        >>> div(Safe("Hello <b>World!</b>")).to_html()
        '<div>Hello <b>World!</b></div>'
    """

    escape = False


class JavaScript(Safe):
    """Marker for a JavaScript string.

    The content of this string is not escaped when rendered.

    Usage:

        js = JavaScript("alert('Hello World!')")
    """


class Attrs(TypedDict, total=False):
    """Attributes of an element or component.

    Example usage::

        class PersonAttrs(Attributes):
            name: str
            age: NotRequired[int]

        class Person(Component[PersonAttrs]):
            @override
            def render(self) -> dl:
                return dl(
                    dt("Name"),
                    dd(self.attrs["name"]),
                    dt("Age"),
                    dd(self.attrs.get("age", "N/A")),
                )
    """


class NoAttrs(TypedDict):
    """Placeholder for element with no attributes."""


class BaseElement(metaclass=ABCMeta):
    html_header: ClassVar[str | None] = None
    html_name: ClassVar[str | None] = None

    always_pair: ClassVar[bool] = False
    formatter: ClassVar[FormatContext] = FormatContext("element_formatter")

    classes: ClassVar[Sequence[str]] = []
    styles: ClassVar["GlobalStyles"] = {}

    children: Sequence[Any]
    attrs: Mapping[str, Any]

    context: dict[str, Any]

    def __init_subclass__(cls) -> None:
        ELEMENT_REGISTRY.setdefault(cls.__name__, [])
        ELEMENT_REGISTRY[cls.__name__].append(cls)

    def __init__(self, *children: Any, **attrs: Any) -> None:
        self.context = {}
        self.children = children
        self.attrs = attrs

    def __str__(self) -> str:
        return self.to_string()

    def __format__(self, _: str) -> str:
        return self.formatter.append(self)

    def __len__(self) -> int:
        return len(self.children)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.children)

    def __repr__(self) -> str:
        return self.to_string(pretty=False)

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) is type(other)
            and self.children == other.children
            and self.attrs == other.attrs
        )

    def _format_attributes(
        self, classes: list[str] | None = None, is_html: bool = False
    ) -> str:
        attrs: dict[str, Any]
        if is_html:
            attrs = format_attrs(type(self), dict(self.attrs), is_html=True)
        else:
            attrs = self.aliased_attrs

        if classes:
            if "class" in attrs:
                attrs["class"] += " " + " ".join(classes)
            else:
                attrs["class"] = " ".join(classes)

        return " ".join(
            f'{key}="{value}"' if '"' not in value else f"{key}='{value}'"
            for key, value in attrs.items()
        )

    def _format_children(
        self,
        format_fun: Callable[[Any], str] = format_element,
    ) -> str:
        formatted = []
        for child in self.children:
            if isinstance(child, BaseElement):
                child.context.update(self.context)
            formatted.append(format_fun(child))
        return "".join(formatted)

    @property
    def aliased_attrs(self) -> dict[str, Any]:
        """Attributes as a dict with keys renamed to their aliases."""
        return format_attrs(type(self), dict(self.attrs))

    @property
    def text(self) -> str:
        """Get the text content of the element."""
        return "".join(
            child.text if isinstance(child, BaseElement) else str(child)
            for child in self.children
        )

    @property
    def theme(self) -> Theme:
        """Get the theme of the element."""
        if context_theme := self.context.get("theme"):
            if isinstance(context_theme, Theme):
                return context_theme
        return get_default_theme()

    def is_simple(self) -> bool:
        """Check if the element is simple (i.e. contains only one primitive type)."""
        return len(self) == 1 and isinstance(self.children[0], str | int | float | bool)

    def has_attributes(self) -> bool:
        """Check if the element has any attributes."""
        return bool(self.attrs)

    def to_string(self, pretty: bool = True, _level: int = 0) -> str:
        """Convert the element tree to a string representation.

        Args:
            pretty (bool, optional): Whether to indent the string. Defaults to True.

        Returns:
            str: The string representation of the element tree.
        """
        indent = "  " * _level if pretty else ""
        name = self.__class__.__name__
        element = f"<{name}"

        if self.has_attributes():
            element += f" {self._format_attributes()}"

        if self.children:
            prefix, sep, suffix = "", "", ""
            if pretty and (not self.is_simple() or self.has_attributes()):
                prefix, sep, suffix = f"\n{indent}  ", f"\n{indent}  ", f"\n{indent}"

            children_str = sep.join(
                child.to_string(pretty=pretty, _level=_level + 1)
                if isinstance(child, BaseElement)
                else str(child)
                for child in self.children
            )

            element += f">{prefix}{children_str}{suffix}</{name}>"
        else:
            element += " />"

        return element

    def to_html(self) -> str:
        """Convert an element tree to an HTML string."""
        dom = self
        classes = list(dom.classes)

        while dom != (rendered_dom := dom.render()):
            rendered_dom.context.update(dom.context)
            dom = rendered_dom
            classes += dom.classes

        element_tag = f"{dom.html_header}\n" if dom.html_header else ""
        children_str = dom._format_children() if dom.children else ""

        if dom.html_name == "__hidden__":
            element_tag += children_str
            return element_tag

        element_tag += f"<{dom.html_name}"
        if dom.has_attributes() or classes:
            attributes_str = dom._format_attributes(classes, is_html=True)
            element_tag += f" {attributes_str}"

        if dom.children or dom.always_pair:
            element_tag += f">{children_str}</{dom.html_name}>"
        else:
            element_tag += " />"

        return element_tag

    def attrs_for(self, cls: type["BaseElement"]) -> dict[str, Any]:
        """Get the attributes of this component that are defined in the given element.

        This is useful so that you can pass common attributes to an element
        without having to pass them from a parent one by one.

        Args:
            cls (type[BaseElement]): The element to get the attributes of.

        """
        return {
            key: value
            for key, value in self.attrs.items()
            if key in get_element_attrs_annotations(cls)
        }

    def render(self) -> "BaseElement":
        return self


NoChildren: TypeAlias = Never
"""Type alias for elements that are not allowed to have children."""

PrimitiveChildren: TypeAlias = str | bool | int | float
"""Type alias for elements that are allowed to have only primitive children.

Primitive children are ``str``, ``bool``, ``int`` and ``float``.
"""

ComplexChildren: TypeAlias = BaseElement
"""Type alias for elements that are allowed to have only non-primitive children."""

AnyChildren: TypeAlias = PrimitiveChildren | ComplexChildren | Safe
"""Type alias for elements that are allowed to have any children."""

TChildren = TypeVar("TChildren", bound=AnyChildren, covariant=True)
"""Type variable for elements representing type of children (the type of *args).

See also: :class:`ludic.types.Component`.
"""

TChildrenArgs = TypeVarTuple("TChildrenArgs")
"""Type variable for strict elements representing type of children (the type of *args).

See also: :class:`ludic.types.ComponentStrict`.
"""

TAttrs = TypeVar("TAttrs", bound=Attrs | NoAttrs, covariant=True)
"""Type variable for elements representing type of attributes (the type of **kwargs)."""


class Element(Generic[TChildren, TAttrs], BaseElement):
    """Base class for Ludic elements.

    Args:
        *children (TChild): The children of the element.
        **attrs (Unpack[TAttrs]): The attributes of the element.
    """

    children: tuple[TChildren, ...]
    attrs: TAttrs

    def __init__(
        self,
        *children: TChildren,
        # FIXME: https://github.com/python/typing/issues/1399
        **attributes: Unpack[TAttrs],  # type: ignore
    ) -> None:
        super().__init__()
        self.attrs = cast(TAttrs, attributes)
        self.children = tuple(self.formatter.extract(*children))


class ElementStrict(Generic[*TChildrenArgs, TAttrs], BaseElement):
    """Base class for strict elements (elements with concrete types of children).

    Args:
        *children (*TChildTuple): The children of the element.
        **attrs (Unpack[TAttrs]): The attributes of the element.
    """

    children: tuple[*TChildrenArgs]
    attrs: TAttrs

    def __init__(
        self,
        *children: *TChildrenArgs,
        # FIXME: https://github.com/python/typing/issues/1399
        **attrs: Unpack[TAttrs],  # type: ignore
    ) -> None:
        super().__init__()
        self.attrs = cast(TAttrs, attrs)
        self.children = tuple(self.formatter.extract(*children))



from collections.abc import Mapping, MutableMapping
from typing import TYPE_CHECKING, Any

from .themes import Theme, get_default_theme
from .types import CSSProperties, GlobalStyles

if TYPE_CHECKING:
    from ludic.types import BaseElement

GLOBAL_STYLES_CACHE: MutableMapping[str, GlobalStyles] = {}


def format_styles(styles: GlobalStyles, separator: str = "\n") -> str:
    """Format styles from all registered elements.

    Args:
        styles (GlobalStyles): Styles to format.
    """
    result: list[str] = []
    nodes_to_parse: list[
        tuple[list[str], str | Mapping[str | tuple[str, ...], Any]]
    ] = [([], styles)]

    while nodes_to_parse:
        parents, node = nodes_to_parse.pop(0)

        content = []
        if isinstance(node, str):
            content.append(node)
        else:
            for key, value in node.items():
                if isinstance(value, str | int | float):
                    content.append(f"{key}: {value};")
                elif isinstance(value, Mapping):
                    keys = (key,) if isinstance(key, str | int | float) else key
                    for key in keys:
                        if key.startswith("@"):
                            value = format_styles(value, separator=" ")
                        nodes_to_parse.append(([*parents, key], value))

        if content:
            result.append(f"{" ".join(parents)} {{ {" ".join(content)} }}")

    return separator.join(result)


def from_components(
    *components: type["BaseElement"], theme: Theme | None = None
) -> GlobalStyles:
    """Global styles collector from given components.

    Example usage:

        class Page(Component[AnyChildren, NoAttrs]):

            @override
            def render(self) -> html:
                return html(
                    head(
                        title("An example Example"),
                        styles(collect_from_components()),
                    ),
                    body(
                        *self.children,
                    ),
                )

    This would render an HTML page containing the ``<style>`` element
    with the styles the given components.
    """
    theme = theme or get_default_theme()
    all_styles: dict[str | tuple[str, ...], CSSProperties | GlobalStyles] = {}

    for component in components:
        styles = getattr(component, "styles", {})

        if not isinstance(styles, Mapping):
            continue
        elif hasattr(styles, "context"):
            styles.context["theme"] = theme

        for key, value in styles.items():
            if isinstance(value, Mapping):
                all_styles[key] = value

    return all_styles


def from_loaded(cache: bool = False, theme: Theme | None = None) -> GlobalStyles:
    """Global styles collector from loaded components.

    Args:
        cache (bool): Whether to cache the result Default is False.

    Returns:
        GlobalStyles: Collected styles from loaded components.
    """
    from ludic.base import ELEMENT_REGISTRY

    global GLOBAL_STYLES_CACHE
    theme = theme or get_default_theme()

    if cache and GLOBAL_STYLES_CACHE.get(theme.name):
        return GLOBAL_STYLES_CACHE[theme.name]

    loaded = (element for elements in ELEMENT_REGISTRY.values() for element in elements)
    result = from_components(*loaded, theme=theme)
    if cache:
        GLOBAL_STYLES_CACHE[theme.name] = result
    return result



import itertools
from collections.abc import Callable
from typing import Any

import starlette
from starlette.datastructures import URL, URLPath
from starlette.routing import NoMatchFound, Route, Router


def join_mounts(prefix: str, suffix: str) -> str:
    """Join mount prefixes and suffixes.

    Args:
        prefix: The prefix to join.
        suffix: The suffix to join.
    """
    prefixes = prefix.split(":")
    suffixes = suffix.split(":")

    for part in prefixes:
        if suffixes and suffixes[0] == part:
            suffixes.pop(0)

    return ":".join(itertools.chain(prefixes, suffixes))


class Request(starlette.requests.Request):
    def url_path_for(
        self, endpoint: str | Callable[..., Any], /, **path_params: Any
    ) -> URLPath:
        """Get URL path for an endpoint.

        Args:
            endpoint: The endpoint to get the URL path for, can be name or a registered
                endpoint class.
            **path_params: URL path parameters.

        Returns:
            The URL path.
        """
        router: Router = self.scope["router"]

        if isinstance(endpoint, str):
            endpoint_name = endpoint
        elif (route := getattr(endpoint, "route", None)) and isinstance(route, Route):
            endpoint_name = route.name
        else:
            raise NoMatchFound(str(endpoint), path_params)

        try:
            return router.url_path_for(endpoint_name, **path_params)
        except NoMatchFound as e:
            if partial_mount := self.scope.get("partial_mount"):
                return router.url_path_for(
                    join_mounts(partial_mount, endpoint_name),
                    **path_params,
                )
            else:
                raise e

    def url_for(self, endpoint: str | Callable[..., Any], /, **path_params: Any) -> URL:
        """Get URL for an endpoint.

        Args:
            endpoint: The endpoint to get the URL for, can be name or a registered
                endpoint class.
            **path_params: URL path parameters.

        Returns:
            The URL.
        """
        url_path = self.url_path_for(endpoint, **path_params)
        return url_path.make_absolute_url(base_url=self.base_url)



from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from typing import Any, Generic, get_args, get_type_hints, override

from starlette.datastructures import FormData
from typeguard import TypeCheckError, check_type

from ludic.types import TAttrs
from ludic.utils import get_annotations_metadata_of_type

from .exceptions import BadRequestError

Parsers = dict[str, Callable[[Any], Any]]


class ValidationError(BadRequestError):
    def __init__(self, detail: str) -> None:
        super().__init__(detail=detail)


class BaseParser(Generic[TAttrs], metaclass=ABCMeta):
    _form_data: FormData
    _parsers: Parsers
    _spec: type[TAttrs]

    @property
    def form_data(self) -> FormData:
        return self._form_data

    @property
    def parsers(self) -> Parsers:
        if not hasattr(self, "_parsers"):
            self._load_meta()
        return self._parsers

    @property
    def spec(self) -> type[TAttrs]:
        if not hasattr(self, "_spec"):
            self._load_meta()
        return self._spec

    def __init__(self, data: FormData, spec: type[TAttrs] | None = None) -> None:
        self._form_data = data
        if spec is not None:
            self._spec = spec

    def _load_meta(self) -> None:
        # This method cannot be part of __init__ as the __orig_class__
        # is not present at that time.
        if (
            not hasattr(self, "_spec")
            and (orig_class := getattr(self, "__orig_class__", None))
            and (bases := get_args(orig_class))
        ):
            self._spec = bases[0]
        else:
            raise TypeError(
                f"Could not collect type information from {self.__class__!r}"
            )

        hints = get_type_hints(self._spec, include_extras=True)
        self._parsers = get_annotations_metadata_of_type(
            hints,
            Callable,  # type: ignore
            default=str,
        )

    def _validate(self, attrs: dict[str, Any]) -> TAttrs:
        try:
            return check_type(attrs, self.spec)
        except TypeCheckError as err:
            raise ValidationError(f"Invalid attributes for {self.spec!r}: {err}.")

    @abstractmethod
    def parse(self) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def validate(self) -> Any:
        raise NotImplementedError()


class Parser(BaseParser[TAttrs]):
    """A class for form data parsing and validation.

    Usage:

        @api.post("/entities")
        def create_entity(data: Validator[EntityAttrs]) -> form:
            attrs = data.validate()
            return form(...)

    Args:
        data (FormData): Form data to be parsed and validated.

    Raises:
        TypeError: The class can raise a TypeError when it was not passible
            to collect type information from generic class.
        ValidationError: The class can raise a ValidationError when parsing
            form data was not successful due to a validation error.
    """

    @override
    def parse(self) -> dict[str, Any]:
        """Parse the attributes and return them.

        The parsers used are defined in the ``Validator._parsers`` dictionary.
        Key is name of the attribute and value is the parser.

        Returns:
            dict[str, Any]: The parsed attributes.
        """
        return {
            key: self.parsers[key](value)
            for key, value in self.form_data.items()
            if key in self.parsers
        }

    @override
    def validate(self) -> TAttrs:
        """Parse, validate and return attributes.

        Returns:
            TAttrs: The validated attributes.

        Raises:
            ValidationError: If the attributes are invalid.
        """
        attrs = self.parse()
        return self._validate(attrs)


class ListParser(BaseParser[TAttrs]):
    @override
    def parse(self) -> list[dict[str, Any]]:
        """Parse a list of attributes and return them.

        The parsers used are defined in the ``Validator._parsers`` dictionary.
        Key is name of the attribute and value is the parser.

        Returns:
            list[dict[str, Any]]: The parsed attributes.
        """
        result: dict[str, dict[str, Any]] = {}
        for compound_key, value in self.form_data.items():
            try:
                key, id_name, id_value = compound_key.split(":", 2)
            except ValueError:
                raise ValidationError(
                    "All keys in a list must contain a unique identifier."
                )
            if key not in self.parsers:
                continue

            result.setdefault(
                id_value,
                {id_name: self.parsers.get(id_name, str)(id_value)}
                if not id_name.startswith("_")
                else {},
            )
            result[id_value][key] = self.parsers[key](value)
        return list(result.values())

    @override
    def validate(self) -> list[TAttrs]:
        """Parse, validate and return attributes in a list.

        Returns:
            list[TAttrs]: The validated attributes.

        Raises:
            ValidationError: If the attributes are invalid.
        """
        result = []
        for attrs in self.parse():
            result.append(self._validate(attrs))
        return result



from collections.abc import Iterable, Mapping
from typing import TypedDict

from .base import (
    AnyChildren,
    Attrs,
    BaseElement,
    ComplexChildren,
    Element,
    ElementStrict,
    JavaScript,
    NoAttrs,
    NoChildren,
    PrimitiveChildren,
    Safe,
    TAttrs,
    TChildren,
    TChildrenArgs,
)
from .components import Blank, Component, ComponentStrict
from .styles import CSSProperties, GlobalStyles, Theme

JSONType = (
    Mapping[str, "JSONType"] | Iterable["JSONType"] | str | int | float | bool | None
)
Headers = Mapping[str, JSONType]
HXHeaders = TypedDict(
    "HXHeaders",
    {
        "HX-Location": JSONType,
        "HX-Push-Url": str,
        "HX-Redirect": str,
        "HX-Refresh": bool,
        "HX-Replace-Url": str,
        "HX-Reswap": str,
        "HX-Retarget": str,
        "HX-Reselect": str,
        "HX-Trigger": JSONType,
        "HX-Trigger-After-Settle": JSONType,
        "HX-Trigger-After-Swap": JSONType,
    },
    total=False,
)

__all__ = (
    "AnyChildren",
    "Attrs",
    "BaseElement",
    "ComplexChildren",
    "Component",
    "ComponentStrict",
    "CSSProperties",
    "Element",
    "ElementStrict",
    "GlobalStyles",
    "Headers",
    "HXHeaders",
    "JavaScript",
    "NoAttrs",
    "NoChildren",
    "PrimitiveChildren",
    "Safe",
    "TAttrs",
    "TChildren",
    "TChildrenArgs",
    "Theme",
    "Blank",
)



from abc import ABCMeta, abstractmethod

from .base import BaseElement, Element, ElementStrict
from .types import NoAttrs, TAttrs, TChildren, TChildrenArgs


class Component(Element[TChildren, TAttrs], metaclass=ABCMeta):
    """Base class for components.

    A component subclasses an :class:`Element` and represents any element
    that can be rendered in Ludic.

    Example usage:

        class PersonAttrs(Attributes):
            age: NotRequired[int]

        class Person(Component[PersonAttrs]):
            @override
            def render(self) -> dl:
                return dl(
                    dt("Name"),
                    dd(self.attrs["name"]),
                    dt("Age"),
                    dd(self.attrs.get("age", "N/A")),
                )

    Now the component can be used in any other component or element:

        >>> div(Person(name="John Doe", age=30), id="person-detail")

    """

    @abstractmethod
    def render(self) -> BaseElement:
        """Render the component as an instance of :class:`BaseElement`."""


class ComponentStrict(ElementStrict[*TChildrenArgs, TAttrs], metaclass=ABCMeta):
    """Base class for strict components.

    A component subclasses an :class:`ElementStrict` and represents any
    element that can be rendered in Ludic. The difference between
    :class:`Component` and :class:`ComponentStrict` is that the latter
    expects concrete types of children. It allows specification
    of each child's type.

    Example usage:

        class PersonAttrs(Attributes):
            age: NotRequired[int]

        class Person(ComponentStrict[str, str, PersonAttrs]):
            @override
            def render(self) -> dl:
                return dl(
                    dt("Name"),
                    dd(" ".join(self.children),
                    dt("Age"),
                    dd(self.attrs.get("age", "N/A")),
                )

    Valid usage would look like this:

        >>> div(Person("John", "Doe", age=30), id="person-detail")

    In this case, we specified that Person expects two string children.
    First child is the first name, and the second is the second name.
    We also specify age as an optional key-word argument.
    """

    @abstractmethod
    def render(self) -> BaseElement:
        """Render the component as an instance of :class:`BaseElement`."""


class Blank(Element[TChildren, NoAttrs]):
    """Element representing no element at all, just children.

    The purpose of this element is to be able to return only children
    when rendering a component.
    """

    html_name = "__hidden__"

    def __init__(self, *children: TChildren) -> None:
        super().__init__(*self.formatter.extract(*children))



from typing import NotRequired, override

try:
    from pygments import highlight
    from pygments.formatters import HtmlFormatter
    from pygments.lexers import get_lexer_by_name

    pygments_loaded = True
except ImportError:
    pygments_loaded = False

from ludic.attrs import GlobalAttrs, HyperlinkAttrs
from ludic.html import a, code, p, pre, style
from ludic.types import (
    AnyChildren,
    Attrs,
    Component,
    ComponentStrict,
    PrimitiveChildren,
    Safe,
)

from .utils import remove_whitespaces


class LinkAttrs(Attrs):
    to: str
    external: NotRequired[bool]
    classes: NotRequired[list[str]]


class Link(ComponentStrict[PrimitiveChildren, LinkAttrs]):
    """Simple component simulating a link.

    The main difference between :class:`Link` and :class:`a` is that this component
    automatically adds the ``target="_blank"`` attribute to the link if the link
    points to an external resource. This can be overridden by setting the
    ``external`` attribute to ``False``.

    Another difference between :class:`Link` and :class:`a` is that this component
    expects only one primitive child, so it cannot contain any nested elements.

    Example usage:

        Link("Hello, World!", to="https://example.com")
    """

    @override
    def render(self) -> a:
        attrs: HyperlinkAttrs = {"href": self.attrs.get("to", "#")}
        if "classes" in self.attrs:
            attrs["classes"] = self.attrs["classes"]

        match self.attrs.get("external", "auto"):
            case True:
                attrs["target"] = "_blank"
            case False:
                pass
            case "auto":
                if str(attrs["href"]).startswith("http"):
                    attrs["target"] = "_blank"

        return a(self.children[0], **attrs)


class Paragraph(Component[AnyChildren, GlobalAttrs]):
    """Simple component simulating a paragraph.

    There is basically just an alias for the :class:`p` element.

    Example usage:

        Paragraph(f"Hello, {b("World")}!")
    """

    @override
    def render(self) -> p:
        return p(*self.children, **self.attrs)


class Code(Component[str, GlobalAttrs]):
    """Simple component simulating a code block.

    Example usage:

        Code("print('Hello, World!')")
    """

    classes = ["code"]
    styles = style.use(
        lambda theme: {
            ".code": {
                "background-color": theme.colors.light,
                "padding": f"{theme.sizes.xxxxs * 0.3} {theme.sizes.xxxs}",
                "border-radius": theme.rounding.less,
                "font-size": theme.fonts.size * 0.9,
            }
        }
    )

    @override
    def render(self) -> code:
        return code(*self.children, **self.attrs)


class CodeBlockAttrs(GlobalAttrs, total=False):
    language: str
    remove_whitespaces: bool


class CodeBlock(Component[str, CodeBlockAttrs]):
    """Simple component simulating a code block.

    Example usage:

        CodeBlock("print('Hello, World!')")
    """

    classes = ["code-block"]
    styles = style.use(
        lambda theme: {
            ".code-block": {
                "background-color": theme.code.background_color,
                "padding-block": theme.sizes.l,
                "padding-inline": theme.sizes.xxl,
                "font-size": theme.fonts.size * 0.9,
            },
        }
    )

    @override
    def render(self) -> pre:
        content = "".join(self.children)

        if self.attrs.get("remove_whitespaces", True):
            content = remove_whitespaces(content)

        if pygments_loaded and (language := self.attrs.get("language")):
            lexer = get_lexer_by_name(language)
            formatter = HtmlFormatter(
                noclasses=True,
                nobackground=True,
                nowrap=True,
                style=self.theme.code.style,
            )
            block = Safe(highlight(content, lexer, formatter))
            return pre(block, **self.attrs_for(pre))
        else:
            return pre(content, **self.attrs_for(pre))



def attr_to_camel(name: str) -> str:
    """Convert an attribute name to camel case.

    Args:
        name (str): The attribute name.

    Returns:
        str: The attribute name in camel case.
    """
    return " ".join(value.title() for value in name.split("_"))


def text_to_kebab(name: str) -> str:
    """Convert a text to kebab case.

    Args:
        name (str): The text.

    Returns:
        str: The text in kebab case.
    """
    return "-".join(value.lower() for value in name.split())


def remove_whitespaces(text: str) -> str:
    """Remove leading whitespaces from a text.

    Args:
        text (str): The text.

    Returns:
        str: The text without leading whitespaces.
    """
    if not text:
        return text

    by_line = text.splitlines()
    min_whitespaces = min(len(line) - len(line.lstrip()) for line in by_line if line)
    return "\n".join(line[min_whitespaces:].rstrip() for line in by_line).strip()
